# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Type definitions - Re-exports from core.types for backwards compatibility.

DEPRECATED: Import from featureform.core.types instead.
"""

# Re-export from core.types for backwards compatibility
from .core.types import VectorType, pd_to_ff_datatype, type_from_proto

__all__ = ["VectorType", "pd_to_ff_datatype", "type_from_proto"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Registration and client functionality for Featureform.

REFACTORING COMPLETE:
---------------------
This module has been refactored into focused packages. All classes have been moved to:

- Provider classes → providers/
- Column resources → registrar/column_resources.py
- Transformation decorators → registrar/transformation_decorators.py
- Registrar class → registrar/registrar.py
- ResourceClient → client_legacy/resource_client.py
- Entity decorator → registrar/entity_decorator.py

This file now serves as a backwards compatibility shim, re-exporting everything
from the new locations to maintain API compatibility.
"""

# Re-export from providers
# Type union for deletable resources
from typing import Union

# Re-export from client
from .client import ResourceClient

# Re-export from core.exceptions
from .core.exceptions import InvalidSQLQuery

# Re-export from enums
from .enums import ScalarType

# Re-export from parse
from .parse import canonicalize_function_definition
from .providers import (
    FileStoreProvider,
    KafkaProvider,
    OfflineK8sProvider,
    OfflineProvider,
    OfflineSparkProvider,
    OfflineSQLProvider,
    OnlineProvider,
)

# Re-export from registrar
from .registrar import (
    ONE_DAY_TARGET_LAG,
    AggregateBuiltFeatures,
    AggregateConfig,
    AttributeBuiltFeature,
    BackfillConfig,
    BuiltFeatures,
    ColumnMapping,
    ColumnResource,
    DFTransformationDecorator,
    EmbeddingColumnResource,
    EntityRegistrar,
    Feature,
    FeatureBuilder,
    FeatureColumnResource,
    FeatureInput,
    FeatureSourceType,
    FeatureType,
    Incremental,
    LabelColumnResource,
    RealtimeBuiltFeature,
    RealtimeFeatureConfig,
    RealtimeInput,
    Registrar,
    ResourceRegistrar,
    SQLTransformationDecorator,
    SubscriptableTransformation,
    Variants,
    entity,
    global_registrar,
)

# Re-export from registrar.registrars
from .registrar.registrars import ColumnSourceRegistrar, SourceRegistrar

# Re-export from resources
from .resources import (
    AggregateFeature,
    AttributeFeature,
    DFTransformation,
    Entity,
    FeatureVariant,
    FeatureView,
    LabelVariant,
    Model,
    PostgresConfig,
    PrimaryData,
    ResourceVariant,
    SnowflakeConfig,
    SourceVariant,
    SQLTable,
    SQLTransformation,
    TrainingSetVariant,
)

# Re-export from types
from .types import VectorType

# Re-export from utils
from .utils.helpers import get_name_variant

DeletableResourceObjects = Union[
    FeatureColumnResource,
    SubscriptableTransformation,
    LabelColumnResource,
    TrainingSetVariant,
    LabelVariant,
    SourceVariant,
    FeatureVariant,
    ColumnSourceRegistrar,
    OnlineProvider,
    OfflineProvider,
    FeatureView,
]

# Global registrar instance and method exports
state = global_registrar.state
clear_state = global_registrar.clear_state
get_state = global_registrar.get_state
set_run = global_registrar.set_run
set_variant_prefix = global_registrar.set_variant_prefix
get_run = global_registrar.get_run
register_user = global_registrar.register_user
register_redis = global_registrar.register_redis
register_pinecone = global_registrar.register_pinecone
register_weaviate = global_registrar.register_weaviate
register_blob_store = global_registrar.register_blob_store
register_bigquery = global_registrar.register_bigquery
register_clickhouse = global_registrar.register_clickhouse
register_firestore = global_registrar.register_firestore
register_cassandra = global_registrar.register_cassandra
register_dynamodb = global_registrar.register_dynamodb
register_mongodb = global_registrar.register_mongodb
register_snowflake = global_registrar.register_snowflake
register_snowflake_legacy = global_registrar.register_snowflake_legacy
register_postgres = global_registrar.register_postgres
register_redshift = global_registrar.register_redshift
register_spark = global_registrar.register_spark
register_k8s = global_registrar.register_k8s
register_kafka = global_registrar.register_kafka
register_vault = global_registrar.register_vault
register_s3 = global_registrar.register_s3
register_hdfs = global_registrar.register_hdfs
register_gcs = global_registrar.register_gcs
register_entity = global_registrar.register_entity
register_column_resources = global_registrar.register_column_resources
register_training_set = global_registrar.register_training_set
register_feature_view = global_registrar.register_feature_view
register_model = global_registrar.register_model
sql_transformation = global_registrar.sql_transformation
register_sql_transformation = global_registrar.register_sql_transformation
get_entity = global_registrar.get_entity
get_source = global_registrar.get_source
get_redis = global_registrar.get_redis
get_postgres = global_registrar.get_postgres
get_dynamodb = global_registrar.get_dynamodb
get_mongodb = global_registrar.get_mongodb
get_snowflake = global_registrar.get_snowflake
get_snowflake_legacy = global_registrar.get_snowflake_legacy
get_redshift = global_registrar.get_redshift
get_bigquery = global_registrar.get_bigquery
get_clickhouse = global_registrar.get_clickhouse
get_spark = global_registrar.get_spark
get_kubernetes = global_registrar.get_kubernetes
get_blob_store = global_registrar.get_blob_store
get_s3 = global_registrar.get_s3
get_gcs = global_registrar.get_gcs
ondemand_feature = global_registrar.ondemand_feature

# ScalarType enum value exports
Nil = ScalarType.NIL
String = ScalarType.STRING
Int = ScalarType.INT
Int32 = ScalarType.INT32
Int64 = ScalarType.INT64
Float32 = ScalarType.FLOAT32
Float64 = ScalarType.FLOAT64
Bool = ScalarType.BOOL
DateTime = ScalarType.DATETIME

__all__ = [
    # Provider classes
    "OfflineProvider",
    "OfflineSQLProvider",
    "OfflineSparkProvider",
    "OfflineK8sProvider",
    "KafkaProvider",
    "OnlineProvider",
    "FileStoreProvider",
    # Column resources
    "ColumnResource",
    "FeatureColumnResource",
    "LabelColumnResource",
    "EmbeddingColumnResource",
    # Transformation decorators
    "SubscriptableTransformation",
    "SQLTransformationDecorator",
    "DFTransformationDecorator",
    # Main classes
    "Registrar",
    "ResourceClient",
    "Incremental",
    "SourceRegistrar",
    "ColumnSourceRegistrar",
    "EntityRegistrar",
    "ResourceRegistrar",
    "ResourceVariant",
    "TrainingSetVariant",
    "Variants",
    "ColumnMapping",
    # Resource classes
    "DFTransformation",
    "SQLTransformation",
    "PrimaryData",
    "SQLTable",
    "Model",
    "Entity",
    "PostgresConfig",
    "SnowflakeConfig",
    "AttributeFeature",
    "AggregateFeature",
    # Feature API v2
    "Feature",
    "FeatureBuilder",
    "FeatureSourceType",
    "FeatureType",
    "AggregateConfig",
    "BackfillConfig",
    "BuiltFeatures",
    "AggregateBuiltFeatures",
    "AttributeBuiltFeature",
    # Realtime Feature API
    "FeatureInput",
    "RealtimeInput",
    "RealtimeBuiltFeature",
    "RealtimeFeatureConfig",
    # Exceptions
    "InvalidSQLQuery",
    # Parse functions
    "canonicalize_function_definition",
    # Utility functions
    "get_name_variant",
    # Decorators
    "entity",
    # Type unions
    "DeletableResourceObjects",
    # Global registrar instance and methods
    "global_registrar",
    "state",
    "clear_state",
    "get_state",
    "set_run",
    "set_variant_prefix",
    "get_run",
    "register_user",
    "register_redis",
    "register_pinecone",
    "register_weaviate",
    "register_blob_store",
    "register_bigquery",
    "register_clickhouse",
    "register_firestore",
    "register_cassandra",
    "register_dynamodb",
    "register_mongodb",
    "register_snowflake",
    "register_snowflake_legacy",
    "register_postgres",
    "register_redshift",
    "register_spark",
    "register_k8s",
    "register_kafka",
    "register_vault",
    "register_s3",
    "register_hdfs",
    "register_gcs",
    "register_entity",
    "register_column_resources",
    "register_training_set",
    "register_feature_view",
    "register_model",
    "sql_transformation",
    "register_sql_transformation",
    "get_entity",
    "get_source",
    "get_redis",
    "get_postgres",
    "get_dynamodb",
    "get_mongodb",
    "get_snowflake",
    "get_snowflake_legacy",
    "get_redshift",
    "get_bigquery",
    "get_clickhouse",
    "get_spark",
    "get_kubernetes",
    "get_blob_store",
    "get_s3",
    "get_gcs",
    "ondemand_feature",
    # Type classes
    "ScalarType",
    "VectorType",
    # ScalarType enum values
    "Nil",
    "String",
    "Int",
    "Int32",
    "Int64",
    "Float32",
    "Float64",
    "Bool",
    "DateTime",
    # Constants
    "ONE_DAY_TARGET_LAG",
]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Catalog configuration classes for data catalogs (Glue, Unity, Snowflake).
"""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional

from typeguard import typechecked

from ..enums import Initialize, RefreshMode, TableFormat
from ..proto import metadata_pb2 as pb


@dataclass
class SnowflakeDynamicTableConfig:
    """
    Configuration for Snowflake dynamic tables.

    Dynamic tables are a Snowflake feature that automatically refreshes data
    based on a target lag and refresh mode.
    """

    target_lag: Optional[str] = None
    refresh_mode: Optional[RefreshMode] = None
    initialize: Optional[Initialize] = None

    def config(self) -> dict:
        return {
            "TargetLag": self.target_lag,
            "RefreshMode": self.refresh_mode.to_string() if self.refresh_mode else None,
            "Initialize": self.initialize.to_string() if self.initialize else None,
        }

    def to_proto(self):
        return pb.SnowflakeDynamicTableConfig(
            target_lag=self.target_lag,
            refresh_mode=self.refresh_mode.to_proto() if self.refresh_mode else None,
            initialize=self.initialize.to_proto() if self.initialize else None,
        )


class Catalog(ABC):
    @abstractmethod
    def type(self):
        pass

    @abstractmethod
    def config(self):
        pass


@typechecked
@dataclass
class GlueCatalog(Catalog):
    region: str
    database: str
    warehouse: str = ""
    assume_role_arn: str = ""
    table_format: TableFormat = field(default_factory=lambda: TableFormat.ICEBERG)

    def __post_init__(self):
        self._validate_database_name()
        self._validate_iceberg_configuration()

    def type(self) -> str:
        return "glue"

    def config(self):
        return {
            "Database": self.database,
            "Warehouse": self.warehouse,
            "Region": self.region,
            "AssumeRoleArn": self.assume_role_arn,
            "TableFormat": self.table_format,
        }

    def _validate_database_name(self):
        if self.database == "":
            raise ValueError("Database name cannot be empty")
        if not all(c.isalnum() or c == "_" for c in self.database):
            raise ValueError("Database name must be alphanumeric and/or underscores")

    def _validate_iceberg_configuration(self):
        if self.table_format is TableFormat.ICEBERG:
            errors = []
            if self.warehouse == "":
                errors.append("warehouse is required for Iceberg tables")
            if self.region == "":
                errors.append("region is required for Iceberg tables")

            if len(errors) > 0:
                raise ValueError(";".join(errors))


@typechecked
@dataclass
class UnityCatalog(Catalog):
    catalog: str
    schema: str
    table_format: TableFormat = field(default_factory=lambda: TableFormat.DELTA)

    def __post_init__(self):
        self._validate_catalog_name()
        self._validate_iceberg_configuration()

    def type(self) -> str:
        return "unity"

    def config(self):
        return {
            "Catalog": self.catalog,
            "Schema": self.schema,
            "TableFormat": self.table_format,
        }

    def _validate_catalog_name(self):
        if self.catalog == "":
            raise ValueError("Database name cannot be empty")
        if self.schema == "":
            raise ValueError("Schema name cannot be empty")

    def _validate_iceberg_configuration(self):
        errors = []
        if self.table_format != TableFormat.DELTA:
            errors.append("table format must be 'DELTA'")

        if len(errors) > 0:
            raise ValueError(";".join(errors))


@typechecked
@dataclass
class SnowflakeCatalog(Catalog):
    external_volume: str
    base_location: str
    table_config: Optional[SnowflakeDynamicTableConfig] = None

    def __post_init__(self):
        if self.table_config is None:
            self.table_config = SnowflakeDynamicTableConfig(
                target_lag="DOWNSTREAM",
                refresh_mode=RefreshMode.AUTO,
                initialize=Initialize.ON_CREATE,
            )
        if not self._validate_target_lag():
            raise ValueError(
                "target_lag must be in the format of '<num> { seconds | minutes | hours | days }' or 'DOWNSTREAM'; the minimum value is 1 minute"
            )

    def type(self) -> str:
        return "snowflake"

    def config(self) -> dict:
        return {
            "ExternalVolume": self.external_volume,
            "BaseLocation": self.base_location,
            "TableFormat": TableFormat.ICEBERG,
            "TableConfig": self.table_config.config(),
        }

    def _validate_target_lag(self) -> bool:
        if self.table_config is None or self.table_config.target_lag is None:
            return True
        # See https://docs.snowflake.com/en/sql-reference/sql/create-dynamic-table#create-dynamic-iceberg-table
        # for more information on the target_lag parameter
        pattern = r"^(\d+)\s+(seconds|minutes|hours|days)$|^(?i:DOWNSTREAM)$"
        match = re.match(pattern, self.table_config.target_lag)

        if not match:
            return False

        if self.table_config.target_lag.lower() == "downstream":
            return True

        value, unit = int(match.group(1)), match.group(2)

        if unit == "seconds" and value < 60:
            return False

        if unit == "minutes" and value < 1:
            return False

        return True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Configuration classes for online store providers.
"""

import json
import re
from dataclasses import dataclass, field
from typing import Optional, Union

from pydantic import BaseModel
from typeguard import typechecked

from .credentials import AWSAssumeRoleCredentials, AWSStaticCredentials, GCPCredentials


@typechecked
@dataclass
class RedisConfig:
    host: str
    port: int
    password: str
    db: int
    ssl_mode: bool = False

    def software(self) -> str:
        return "redis"

    def type(self) -> str:
        return "REDIS_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Addr": f"{self.host}:{self.port}",
            "Password": self.password,
            "DB": self.db,
            "SSLMode": self.ssl_mode,
        }
        return bytes(json.dumps(config), "utf-8")

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, RedisConfig):
            return False
        return (
            self.host == __value.host
            and self.port == __value.port
            and self.password == __value.password
            and self.db == __value.db
            and self.ssl_mode == __value.ssl_mode
        )


@typechecked
@dataclass
class PineconeConfig:
    project_id: str = ""
    environment: str = ""
    api_key: str = ""

    def software(self) -> str:
        return "pinecone"

    def type(self) -> str:
        return "PINECONE_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "ProjectID": self.project_id,
            "Environment": self.environment,
            "ApiKey": self.api_key,
        }
        return bytes(json.dumps(config), "utf-8")

    def deserialize(self, config):
        config = json.loads(config)
        self.project_id = config["ProjectID"]
        self.environment = config["Environment"]
        self.api_key = config["ApiKey"]
        return self

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, PineconeConfig):
            return False
        return (
            self.project_id == __value.project_id
            and self.environment == __value.environment
            and self.api_key == __value.api_key
        )


class WeaviateConfig(BaseModel):
    url: str = ""
    api_key: str = ""

    def software(self) -> str:
        return "weaviate"

    def type(self) -> str:
        return "WEAVIATE_ONLINE"

    def serialize(self) -> bytes:
        if self.url == "":
            raise Exception("URL cannot be empty")
        config = {
            "URL": self.url,
            "ApiKey": self.api_key,
        }
        return bytes(json.dumps(config), "utf-8")

    @classmethod
    def deserialize(cls, config: Union[str, bytes]) -> "WeaviateConfig":
        data = json.loads(config)
        return cls(url=data["URL"], api_key=data["ApiKey"])


@typechecked
@dataclass
class FirestoreConfig:
    collection: str
    project_id: str
    credentials: GCPCredentials

    def software(self) -> str:
        return "firestore"

    def type(self) -> str:
        return "FIRESTORE_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Collection": self.collection,
            "ProjectID": self.project_id,
            "Credentials": self.credentials.to_json(),
        }
        return bytes(json.dumps(config), "utf-8")

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, FirestoreConfig):
            return False
        return (
            self.collection == __value.collection
            and self.project_id == __value.project_id
        )


@typechecked
@dataclass
class CassandraConfig:
    keyspace: str
    host: str
    port: int
    username: str
    password: str
    consistency: str
    replication: int

    def software(self) -> str:
        return "cassandra"

    def type(self) -> str:
        return "CASSANDRA_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Keyspace": self.keyspace,
            "Addr": f"{self.host}:{self.port}",
            "Username": self.username,
            "Password": self.password,
            "Consistency": self.consistency,
            "Replication": self.replication,
        }
        return bytes(json.dumps(config), "utf-8")

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, CassandraConfig):
            return False
        return (
            self.keyspace == __value.keyspace
            and self.host == __value.host
            and self.port == __value.port
            and self.username == __value.username
            and self.password == __value.password
            and self.consistency == __value.consistency
            and self.replication == __value.replication
        )


@typechecked
@dataclass
class DynamodbConfig:
    region: str
    credentials: Union[AWSStaticCredentials, AWSAssumeRoleCredentials]
    table_tags: Optional[dict] = field(default_factory=dict)

    def __post_init__(self):
        errors = self.validate_table_tags(self.table_tags)
        if len(errors) > 0:
            raise ValueError(
                "Table tags validation errors found:\n" + "\n".join(errors)
            )

    def software(self) -> str:
        return "dynamodb"

    def type(self) -> str:
        return "DYNAMODB_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Region": self.region,
            "Credentials": self.credentials.config(),
            "Tags": self.table_tags,
        }
        return bytes(json.dumps(config), "utf-8")

    @classmethod
    def deserialize(cls, json_bytes: bytes) -> "DynamodbConfig":
        deserialized_config = json.loads(json_bytes.decode("utf-8"))

        try:
            if (
                deserialized_config["Credentials"]["Type"]
                == AWSStaticCredentials.type()
            ):
                credentials = AWSStaticCredentials(
                    deserialized_config["Credentials"]["AccessKeyId"],
                    deserialized_config["Credentials"]["SecretKey"],
                )
            elif (
                deserialized_config["Credentials"]["Type"]
                == AWSAssumeRoleCredentials.type()
            ):
                credentials = AWSAssumeRoleCredentials()
            else:
                raise ValueError("Invalid Credentials Type")

            return DynamodbConfig(
                region=deserialized_config["Region"],
                credentials=credentials,
                table_tags=deserialized_config.get("Tags"),
            )
        except KeyError as e:
            raise ValueError(f"Missing key in deserialized config: {e}")

    def validate_table_tags(self, table_tags: Optional[dict]) -> list:
        """
        Validates that:
        - Keys are at most 128 Unicode characters.
        - Values are at most 256 Unicode characters.
        - Keys and values contain only allowed characters.
        - Keys do not start with "aws".

        Returns a list of validation errors. If the list is empty, the dictionary is valid.
        """
        if table_tags is None:
            return []
        allowed_pattern = re.compile(r"^[a-zA-Z0-9\s+\-=._:/]+$")
        errors = []

        for key, value in table_tags.items():
            if not isinstance(key, str) or not isinstance(value, str):
                errors.append(
                    f"Both keys and values must be strings. Invalid pair: ({key}, {value})"
                )
                continue  # Skip further validation for this key-value pair

            if len(key) > 128:
                errors.append(
                    f"Key '{key}' exceeds the maximum length of 128 characters."
                )

            if len(value) > 256:
                errors.append(
                    f"Value for key '{key}' exceeds the maximum length of 256 characters."
                )

            if not allowed_pattern.match(key):
                errors.append(f"Key '{key}' contains invalid characters.")

            if not allowed_pattern.match(value):
                errors.append(f"Value '{value}' contains invalid characters.")

            if key.startswith("aws"):
                errors.append(f"Key '{key}' cannot start with 'aws'.")

        return errors


@typechecked
@dataclass
class MongoDBConfig:
    username: str
    password: str
    host: str
    port: str
    database: str
    throughput: int

    def software(self) -> str:
        return "mongodb"

    def type(self) -> str:
        return "MONGODB_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Username": self.username,
            "Password": self.password,
            "Host": self.host,
            "Port": self.port,
            "Database": self.database,
            "Throughput": self.throughput,
        }
        return bytes(json.dumps(config), "utf-8")

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, MongoDBConfig):
            return False
        return (
            self.username == __value.username
            and self.password == __value.password
            and self.host == __value.host
            and self.port == __value.port
            and self.database == __value.database
            and self.throughput == __value.throughput
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Core module for Featureform.

This module contains core types, state management, protocols, and exceptions.
"""

from .exceptions import (
    FeatureNotFound,
    InvalidSQLQuery,
    InvalidTrainingSetFeatureComputationMode,
    LabelNotFound,
)
from .protocols import HasNameVariant, ResourceProtocol, ResourceVariant
from .state import ResourceRedefinedError, ResourceState
from .types import VectorType, pd_to_ff_datatype, type_from_proto

__all__ = [
    # Types
    "VectorType",
    "pd_to_ff_datatype",
    "type_from_proto",
    # State
    "ResourceState",
    "ResourceRedefinedError",
    # Protocols
    "HasNameVariant",
    "ResourceProtocol",
    "ResourceVariant",
    # Exceptions
    "FeatureNotFound",
    "LabelNotFound",
    "InvalidSQLQuery",
    "InvalidTrainingSetFeatureComputationMode",
]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Configuration classes for compute providers (Spark, K8s).
"""

import json
from dataclasses import dataclass, field
from typing import Dict, Optional, Union

from typeguard import typechecked

from ..proto import metadata_pb2 as pb


@dataclass
class SparkFlags:
    spark_params: Dict[str, str] = field(default_factory=dict)
    write_options: Dict[str, str] = field(default_factory=dict)
    table_properties: Dict[str, str] = field(default_factory=dict)

    def serialize(self) -> dict:
        return {
            "SparkParams": self.spark_params,
            "WriteOptions": self.write_options,
            "TableProperties": self.table_properties,
        }

    @classmethod
    def deserialize(cls, config: dict) -> Optional["SparkFlags"]:
        """
        Deserialize a dictionary into a SparkFlags object.

        Args:
            config (dict): The dictionary containing SparkFlags configuration.

        Returns:
            Optional[SparkFlags]: The deserialized SparkFlags object, or None if the config is missing.
        """
        if not config:
            return None
        return cls(
            spark_params=config.get("SparkParams", {}),
            write_options=config.get("WriteOptions", {}),
            table_properties=config.get("TableProperties", {}),
        )

    def to_proto(self) -> pb.SparkFlags:
        return pb.SparkFlags(
            spark_params=[
                pb.SparkParam(key=k, value=v) for k, v in self.spark_params.items()
            ],
            write_options=[
                pb.WriteOption(key=k, value=v) for k, v in self.write_options.items()
            ],
            table_properties=[
                pb.TableProperty(key=k, value=v)
                for k, v in self.table_properties.items()
            ],
        )


EmptySparkFlags = SparkFlags()


@typechecked
@dataclass
class SparkConfig:
    executor_type: str
    executor_config: dict
    store_type: str
    store_config: dict
    kafka_config: Optional[dict] = None
    catalog_type: Optional[str] = None
    catalog_config: Optional[dict] = None
    spark_flags: Optional[SparkFlags] = field(default_factory=lambda: EmptySparkFlags)

    def __post_init__(self):
        if self.spark_flags is None:
            self.spark_flags = EmptySparkFlags

    def software(self) -> str:
        return "spark"

    def type(self) -> str:
        return "SPARK_OFFLINE"

    def serialize(self) -> bytes:
        config = {
            "ExecutorType": self.executor_type,
            "StoreType": self.store_type,
            "ExecutorConfig": self.executor_config,
            "StoreConfig": self.store_config,
            "SparkFlags": self.spark_flags.serialize(),
        }

        if self.catalog_type is not None and self.catalog_config is not None:
            config["CatalogType"] = self.catalog_type
            config["CatalogConfig"] = self.catalog_config

        if self.kafka_config is not None:
            config["KafkaConfig"] = self.kafka_config

        return bytes(json.dumps(config), "utf-8")

    @classmethod
    def deserialize(cls, json_bytes: bytes) -> "SparkConfig":
        deserialized_config = json.loads(json_bytes.decode("utf-8"))

        spark_flags = (
            SparkFlags.deserialize(deserialized_config.get("SparkFlags"))
            if deserialized_config.get("SparkFlags")
            else EmptySparkFlags
        )

        try:
            return cls(
                executor_type=deserialized_config["ExecutorType"],
                executor_config=deserialized_config["ExecutorConfig"],
                store_type=deserialized_config["StoreType"],
                store_config=deserialized_config["StoreConfig"],
                catalog_type=deserialized_config.get("CatalogType"),
                catalog_config=deserialized_config.get("CatalogConfig"),
                spark_flags=spark_flags,
            )
        except KeyError as e:
            raise ValueError(f"Missing expected config key: {e}")


@typechecked
@dataclass
class K8sResourceSpecs:
    cpu_request: str = ""
    cpu_limit: str = ""
    memory_request: str = ""
    memory_limit: str = ""


@typechecked
@dataclass
class K8sArgs:
    # TODO Delete and Deprecate
    docker_image: str
    specs: Union[K8sResourceSpecs, None] = None

    def apply(self, transformation: pb.Transformation):
        transformation.kubernetes_args.docker_image = self.docker_image
        if self.specs is not None:
            transformation.kubernetes_args.specs.cpu_request = self.specs.cpu_request
            transformation.kubernetes_args.specs.cpu_limit = self.specs.cpu_limit
            transformation.kubernetes_args.specs.memory_request = (
                self.specs.memory_request
            )
            transformation.kubernetes_args.specs.memory_limit = self.specs.memory_limit
        return transformation


@typechecked
@dataclass
class K8sConfig:
    store_type: str
    store_config: dict
    docker_image: str = ""

    def software(self) -> str:
        return "k8s"

    def type(self) -> str:
        return "K8S_OFFLINE"

    def serialize(self) -> bytes:
        config = {
            "ExecutorType": "K8S",
            "ExecutorConfig": {"docker_image": self.docker_image},
            "StoreType": self.store_type,
            "StoreConfig": self.store_config,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class EmptyConfig:
    def software(self) -> str:
        return ""

    def type(self) -> str:
        return ""

    def serialize(self) -> bytes:
        return bytes("", "utf-8")

    def deserialize(self, config):
        return self



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Model resource classes for Featureform.

This module contains classes for defining and managing models.
"""

from dataclasses import dataclass, field
from typing import Tuple

from typeguard import typechecked

from ..enums import OperationType, ResourceType
from ..proto import metadata_pb2 as pb
from .provider import Properties


@typechecked
@dataclass
class Model:
    name: str
    description: str = ""
    tags: list = field(default_factory=list)
    properties: dict = field(default_factory=dict)

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.MODEL

    def _create(self, req_id, stub) -> Tuple[None, None]:
        properties = pb.Properties(property=self.properties)
        serialized = pb.ModelRequest(
            model=pb.Model(
                name=self.name,
                tags=pb.Tags(tag=self.tags),
                properties=Properties(self.properties).serialized,
            ),
            request_id="",
        )

        stub.CreateModel(serialized)
        return None, None

    def to_dictionary(self):
        return {
            "name": self.name,
            "description": self.description,
            "tags": self.tags,
            "properties": self.properties,
        }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Entity decorator for Featureform.

This module contains the @entity decorator for registering entities and their
associated features and labels.
"""

from .column_resources import (
    EmbeddingColumnResource,
    FeatureColumnResource,
    LabelColumnResource,
)
from .feature_api import FeatureBuilder
from .variants import Variants

__all__ = ["entity"]


# Import register_entity from the module that will contain Registrar
def register_entity(name: str):
    """Register an entity. This will be replaced with actual implementation."""
    from . import global_registrar

    return global_registrar.register_entity(name)


def entity(cls):
    """
    Class decorator for registering entities and their associated features and labels.

    **Examples**
    ```python
    @ff.entity
    class User:
        avg_transactions = ff.Feature()
        fraudulent = ff.Label()

        # Using the new builder pattern
        account_balance = (
            ff.Feature()
            .from_dataset(
                batch_features,
                entity="CustomerID",
                values="Balance",
                timestamp="Timestamp",
            )
        )
    ```

    Returns:
        entity (EntityRegistrar): Decorated entity registrar with features/labels as attributes
    """
    from . import global_registrar

    # 1. Use the lowercase name of the class as the entity name
    entity_obj = register_entity(cls.__name__.lower())
    # 2. Given the Feature/Label/Variant class constructors are evaluated
    #    before the entity decorator, apply the entity name to their
    #    respective name dictionaries prior to registration
    for attr_name in cls.__dict__:
        attr_value = cls.__dict__[attr_name]

        if isinstance(
            attr_value,
            (FeatureColumnResource, LabelColumnResource, EmbeddingColumnResource),
        ):
            resource = attr_value
            resource.name = attr_name if resource.name == "" else resource.name
            resource.entity = entity_obj
            resource.register()
            # Set the resource as an attribute on the entity object
            setattr(entity_obj, attr_name, resource)

        elif isinstance(attr_value, FeatureBuilder):
            # Handle the new builder pattern
            builder = attr_value
            # Set base_name from attribute name if not explicitly set
            if builder.base_name is None:
                builder.base_name = attr_name
            # Set initial_variant from global run if not set
            if builder.initial_variant is None:
                builder.initial_variant = global_registrar.get_run()
            # Normalize values before build() so build() is a pure function
            builder.entity = global_registrar._normalize_entity_reference(entity_obj)
            builder.owner = (
                builder.owner
                if builder.owner
                else global_registrar.must_get_default_owner()
            )
            builder.value_type = global_registrar._normalize_value_type(
                builder.value_type
            )
            # register_feature_builder returns BuiltFeatures containing FeatureVariants
            built_features = global_registrar.register_feature_builder(builder)
            # Set the BuiltFeatures as an attribute on the entity object
            # This allows User.feature_name to return the BuiltFeatures wrapper
            setattr(entity_obj, attr_name, built_features)

        elif isinstance(attr_value, Variants):
            variants = attr_value
            for variant_key, resource in variants.resources.items():
                resource.name = attr_name if resource.name == "" else resource.name
                resource.entity = entity_obj
                resource.register()
            # Set the variants object as an attribute on the entity object
            setattr(entity_obj, attr_name, variants)

    # Return the entity object so it can be used in the code
    return entity_obj



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Registrar module for Featureform.

This module contains registrar classes that wrap registered resources and
provide registration functionality.
"""

from .column_mapping import ColumnMapping
from .column_resources import (
    ColumnResource,
    EmbeddingColumnResource,
    FeatureColumnResource,
    LabelColumnResource,
)
from .entity_decorator import entity
from .feature_api import (
    AggregateBuiltFeatures,
    AggregateConfig,
    AttributeBuiltFeature,
    BackfillConfig,
    BuiltFeatures,
    Feature,
    FeatureBuilder,
    FeatureSourceType,
    FeatureType,
)
from .input_validators import (
    is_streaming_input,
    validate_batch_transformation_inputs,
    validate_streaming_transformation_inputs,
)
from .realtime_feature import (
    FeatureInput,
    RealtimeBuiltFeature,
    RealtimeFeatureConfig,
    RealtimeInput,
)
from .registrar import ONE_DAY_TARGET_LAG, Incremental, Registrar
from .registrars import (
    ColumnSourceRegistrar,
    EntityRegistrar,
    ModelRegistrar,
    ResourceRegistrar,
    SourceRegistrar,
    UserRegistrar,
)
from .transformation_decorators import (
    DFTransformationDecorator,
    SQLTransformationDecorator,
    SubscriptableTransformation,
)
from .variants import Variants

__all__ = [
    "Registrar",
    "Incremental",
    "ONE_DAY_TARGET_LAG",
    "EntityRegistrar",
    "UserRegistrar",
    "SourceRegistrar",
    "ColumnSourceRegistrar",
    "ResourceRegistrar",
    "ModelRegistrar",
    "ColumnMapping",
    "Variants",
    "SubscriptableTransformation",
    "SQLTransformationDecorator",
    "DFTransformationDecorator",
    "ColumnResource",
    "FeatureColumnResource",
    "LabelColumnResource",
    "EmbeddingColumnResource",
    "entity",
    # Feature API v2
    "Feature",
    "FeatureBuilder",
    "FeatureSourceType",
    "FeatureType",
    "AggregateConfig",
    "BackfillConfig",
    "BuiltFeatures",
    "AggregateBuiltFeatures",
    "AttributeBuiltFeature",
    # Input validators
    "is_streaming_input",
    "validate_streaming_transformation_inputs",
    "validate_batch_transformation_inputs",
    # Realtime Feature API
    "FeatureInput",
    "RealtimeInput",
    "RealtimeBuiltFeature",
    "RealtimeFeatureConfig",
]

# Create global registrar instance
global_registrar = Registrar()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Source resource classes for Featureform.

This module contains classes for defining and managing data sources.
"""

from dataclasses import dataclass
from datetime import timedelta
from typing import List, Optional, Tuple, Union

import grpc
from google.protobuf.duration_pb2 import Duration
from typeguard import typechecked

from ..core import ResourceVariant
from ..enums import OperationType, ResourceStatus, ResourceType, SourceType
from ..operations.equivalence import _get_and_set_equivalent_variant
from ..proto import metadata_pb2 as pb
from .base import NameVariant
from .entity import Entity
from .provider import Properties, Provider, ServerStatus
from .schedule import Schedule
from .transformations import PrimaryData, Transformation

# Type alias for source definitions
SourceDefinition = Union[PrimaryData, Transformation, str]


@typechecked
@dataclass
class Source:
    """High-level source resource containing multiple variants."""

    name: str
    default_variant: str
    variants: List[str]

    def to_dictionary(self):
        return {
            "name": self.name,
            "default_variant": self.default_variant,
            "variants": self.variants,
        }


@typechecked
@dataclass
class SourceVariant(ResourceVariant):
    """A specific variant of a source with its definition and metadata."""

    name: str
    definition: SourceDefinition
    owner: str
    provider: str
    description: str
    tags: list
    properties: dict
    variant: str
    created: str = None
    status: str = (
        "ready"  # there is no associated status by default but it always stores ready
    )
    schedule: str = ""
    schedule_obj: Schedule = None
    is_transformation = (
        SourceType.PRIMARY_SOURCE.value
    )  # TODO this is the same as source_type below; pick one
    source_text: str = ""
    source_type: str = ""
    transformation: str = ""
    inputs: list = ([],)
    error: Optional[str] = None
    server_status: Optional[ServerStatus] = None
    max_job_duration: timedelta = timedelta(hours=48)

    def update_schedule(self, schedule) -> None:
        self.schedule_obj = Schedule(
            name=self.name,
            variant=self.variant,
            resource_type=7,
            schedule_string=schedule,
        )
        self.schedule = schedule

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.SOURCE_VARIANT

    def name_variant(self) -> NameVariant:
        return (self.name, self.variant)

    def get(self, stub):
        return SourceVariant.get_by_name_variant(stub, self.name, self.variant)

    @staticmethod
    def get_by_name_variant(stub, name, variant):
        name_variant = pb.NameVariantRequest(
            name_variant=pb.NameVariant(name=name, variant=variant)
        )
        source = next(stub.GetSourceVariants(iter([name_variant])))
        definition = SourceVariant._get_source_definition(source)

        return SourceVariant(
            created=None,
            name=source.name,
            definition=definition,
            owner=source.owner,
            provider=source.provider,
            description=source.description,
            variant=source.variant,
            tags=list(source.tags.tag),
            properties=dict(source.properties.property.items()),
            status=source.status.Status._enum_type.values[source.status.status].name,
            error=source.status.error_message,
            server_status=ServerStatus.from_proto(source.status),
            max_job_duration=source.max_job_duration.ToTimedelta(),
        )

    @staticmethod
    def _get_source_definition(source):
        definition_type = source.WhichOneof("definition")
        if definition_type == "primaryData":
            return PrimaryData.from_proto(source.primaryData)
        elif definition_type == "transformation":
            return Transformation.from_proto(source.transformation)
        else:
            raise Exception(f"Invalid source definition type {definition_type}")

    def _get_and_set_equivalent_variant(self, req_id, stub):
        defArgs = self.definition.kwargs()
        duration = Duration()
        duration.FromTimedelta(self.max_job_duration)

        serialized = pb.SourceVariantRequest(
            source_variant=pb.SourceVariant(
                created=None,
                name=self.name,
                variant=self.variant,
                owner=self.owner,
                description=self.description,
                schedule=self.schedule,
                provider=self.provider,
                tags=pb.Tags(tag=self.tags),
                properties=Properties(self.properties).serialized,
                status=pb.ResourceStatus(status=pb.ResourceStatus.NO_STATUS),
                max_job_duration=duration,
                **defArgs,
            ),
            request_id="",
        )

        existing_variant = _get_and_set_equivalent_variant(
            req_id, serialized, "source_variant", stub
        )
        return serialized, existing_variant, "source_variant"

    def _create(self, req_id, stub) -> Tuple[Optional[str], Optional[str]]:
        serialized, existing_variant, _ = self._get_and_set_equivalent_variant(
            req_id, stub
        )
        if existing_variant is None:
            stub.CreateSourceVariant(serialized)
        return serialized.source_variant.variant, existing_variant

    def get_status(self):
        return ResourceStatus(self.status)

    def is_transformation_type(self):
        return isinstance(self.definition, Transformation)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value


class EntityReference:
    """Reference to an entity resource."""

    name: str
    obj: Union[Entity, None]

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.GET

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.ENTITY

    def _get(self, stub):
        entityList = stub.GetEntities(
            iter([pb.NameRequest(name=pb.Name(name=self.name))])
        )
        try:
            for entity in entityList:
                self.obj = entity
        except grpc._channel._MultiThreadedRendezvous:
            raise ValueError(f"Entity {self.name} not found.")


@typechecked
@dataclass
class ProviderReference:
    """Reference to a provider resource."""

    name: str
    provider_type: str
    obj: Union[Provider, None]

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.GET

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.PROVIDER

    def _get(self, stub):
        providerList = stub.GetProviders(
            iter([pb.NameRequest(name=pb.Name(name=self.name))])
        )
        try:
            for provider in providerList:
                self.obj = provider
        except grpc._channel._MultiThreadedRendezvous:
            raise ValueError(
                f"Provider {self.name} of type {self.provider_type} not found."
            )


@typechecked
@dataclass
class SourceReference:
    """Reference to a source variant resource."""

    name: str
    variant: str
    obj: Union["SourceVariant", None]

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.GET

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.SOURCE_VARIANT

    def name_variant(self) -> NameVariant:
        return (self.name, self.variant)

    def _get(self, stub):
        sourceList = stub.GetSourceVariants(
            iter(
                [
                    pb.NameVariantRequest(
                        name_variant=pb.NameVariant(
                            name=self.name, variant=self.variant
                        )
                    )
                ]
            )
        )
        try:
            for source in sourceList:
                self.obj = source
        except grpc._channel._MultiThreadedRendezvous:
            raise ValueError(f"Source {self.name}, variant {self.variant} not found.")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Label resource classes for Featureform.

This module contains classes for defining and managing labels.
"""

from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, Union

from typeguard import typechecked

from ..config import ResourceSnowflakeConfig
from ..core import ResourceVariant
from ..enums import OperationType, ResourceStatus, ResourceType, ScalarType
from ..operations.equivalence import _get_and_set_equivalent_variant
from ..proto import metadata_pb2 as pb
from ..types import VectorType, type_from_proto
from .base import NameVariant
from .mappings import EntityMapping, EntityMappings, ResourceColumnMapping
from .provider import Properties, ServerStatus

# Type alias for resource location
ResourceLocation = ResourceColumnMapping


@typechecked
@dataclass
class Label:
    name: str
    default_variant: str
    variants: List[str]

    def to_dictionary(self):
        return {
            "name": self.name,
            "default_variant": self.default_variant,
            "variants": self.variants,
        }


@typechecked
@dataclass
class LabelVariant(ResourceVariant):
    name: str
    source: Any
    value_type: Union[VectorType, ScalarType, str]
    entity: str
    owner: str
    description: str
    location: Union[ResourceLocation, EntityMappings]
    variant: str
    tags: Optional[list] = None
    properties: Optional[dict] = None
    provider: Optional[str] = None
    created: str = None
    status: str = "NO_STATUS"
    error: Optional[str] = None
    server_status: Optional[ServerStatus] = None
    resource_snowflake_config: Optional[ResourceSnowflakeConfig] = None

    def __post_init__(self):
        if isinstance(self.value_type, str):
            self.value_type = ScalarType(self.value_type)

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.LABEL_VARIANT

    def name_variant(self) -> NameVariant:
        return (self.name, self.variant)

    def get(self, stub) -> "LabelVariant":
        return LabelVariant.get_by_name_variant(stub, self.name, self.variant)

    @staticmethod
    def get_by_name_variant(stub, name, variant):
        name_variant = pb.NameVariantRequest(
            name_variant=pb.NameVariant(name=name, variant=variant)
        )
        label = next(stub.GetLabelVariants(iter([name_variant])))

        loc_type = label.WhichOneof("location")
        location = None
        if loc_type == "columns":
            location = ResourceColumnMapping(
                entity=label.columns.entity,
                value=label.columns.value,
                timestamp=label.columns.ts,
            )
        elif loc_type == "entity_mappings":
            location = EntityMappings(
                mappings=[
                    EntityMapping(m.name, m.entity_column)
                    for m in label.entity_mappings.mappings
                ],
                value_column=label.entity_mappings.value_column,
                timestamp_column=(
                    label.entity_mappings.timestamp_column
                    if label.entity_mappings.timestamp_column
                    else None
                ),
            )

        return LabelVariant(
            name=label.name,
            variant=label.variant,
            source=(label.source.name, label.source.variant),
            value_type=type_from_proto(label.type),
            entity=label.entity,
            owner=label.owner,
            provider=label.provider,
            location=location,
            description=label.description,
            tags=list(label.tags.tag),
            properties=dict(label.properties.property.items()),
            status=label.status.Status._enum_type.values[label.status.status].name,
            server_status=ServerStatus.from_proto(label.status),
            error=label.status.error_message,
        )

    def _get_and_set_equivalent_variant(self, req_id, stub):
        if hasattr(self.source, "name_variant"):
            self.source = self.source.name_variant()
        label_variant = pb.LabelVariant(
            name=self.name,
            variant=self.variant,
            source=pb.NameVariant(
                name=self.source[0],
                variant=self.source[1],
            ),
            provider=self.provider,
            type=self.value_type.to_proto(),
            entity=self.entity,
            owner=self.owner,
            description=self.description,
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
            status=pb.ResourceStatus(status=pb.ResourceStatus.NO_STATUS),
            resource_snowflake_config=(
                self.resource_snowflake_config.to_proto()
                if self.resource_snowflake_config
                else None
            ),
        )
        if isinstance(self.location, ResourceLocation):
            label_variant.entity_mappings.CopyFrom(
                self.location.to_entity_mappings_proto(self.entity)
            )
        elif isinstance(self.location, EntityMappings):
            label_variant.entity_mappings.CopyFrom(self.location.to_proto())
        else:
            raise ValueError(
                f"Invalid location type {type(self.location)} for LabelVariant {self.name}"
            )
        serialized = pb.LabelVariantRequest(
            label_variant=label_variant,
            request_id="",
        )

        return (
            serialized,
            _get_and_set_equivalent_variant(req_id, serialized, "label_variant", stub),
            "label_variant",
        )

    def _create(self, req_id, stub) -> Tuple[Optional[str], Optional[str]]:
        serialized, existing_variant, _ = self._get_and_set_equivalent_variant(
            req_id, stub
        )
        if existing_variant is None:
            stub.CreateLabelVariant(serialized)
        return serialized.label_variant.variant, existing_variant

    def get_status(self):
        return ResourceStatus(self.status)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Column mapping for registrar.

This module contains the ColumnMapping class used for defining feature and label columns.
"""

from typing import List, Optional

from ..config.offline_stores import ResourceSnowflakeConfig
from ..enums import ResourceType


class ColumnMapping(dict):
    """
    Dictionary-based class for mapping columns to features or labels.

    Attributes:
        name: Name of the feature or label
        column: Column name in the source
        resource_type: Type of resource (FEATURE or LABEL)
        tags: List of tags
        properties: Dictionary of properties
        variant: Variant name (optional)
        resource_snowflake_config: Snowflake-specific configuration (optional)
    """

    name: str
    column: str
    resource_type: ResourceType
    tags: List[str]
    properties: dict
    variant: str = ""
    resource_snowflake_config: Optional[ResourceSnowflakeConfig] = None



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from featureform.proto import serving_pb2 as featureform_dot_proto_dot_serving__pb2

GRPC_GENERATED_VERSION = '1.75.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in featureform/proto/serving_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class FeatureStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TrainingData = channel.unary_stream(
                '/featureform.serving.proto.Feature/TrainingData',
                request_serializer=featureform_dot_proto_dot_serving__pb2.TrainingDataRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.TrainingDataRows.FromString,
                _registered_method=True)
        self.TrainTestSplit = channel.stream_stream(
                '/featureform.serving.proto.Feature/TrainTestSplit',
                request_serializer=featureform_dot_proto_dot_serving__pb2.TrainTestSplitRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.BatchTrainTestSplitResponse.FromString,
                _registered_method=True)
        self.TrainingDataColumns = channel.unary_unary(
                '/featureform.serving.proto.Feature/TrainingDataColumns',
                request_serializer=featureform_dot_proto_dot_serving__pb2.TrainingDataColumnsRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.TrainingColumns.FromString,
                _registered_method=True)
        self.FeatureServe = channel.unary_unary(
                '/featureform.serving.proto.Feature/FeatureServe',
                request_serializer=featureform_dot_proto_dot_serving__pb2.FeatureServeRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.FeatureRow.FromString,
                _registered_method=True)
        self.SourceData = channel.unary_stream(
                '/featureform.serving.proto.Feature/SourceData',
                request_serializer=featureform_dot_proto_dot_serving__pb2.SourceDataRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.SourceDataRows.FromString,
                _registered_method=True)
        self.SourceColumns = channel.unary_unary(
                '/featureform.serving.proto.Feature/SourceColumns',
                request_serializer=featureform_dot_proto_dot_serving__pb2.SourceColumnRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.SourceDataColumns.FromString,
                _registered_method=True)
        self.Nearest = channel.unary_unary(
                '/featureform.serving.proto.Feature/Nearest',
                request_serializer=featureform_dot_proto_dot_serving__pb2.NearestRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.NearestResponse.FromString,
                _registered_method=True)
        self.BatchFeatureServe = channel.unary_stream(
                '/featureform.serving.proto.Feature/BatchFeatureServe',
                request_serializer=featureform_dot_proto_dot_serving__pb2.BatchFeatureServeRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.BatchFeatureRows.FromString,
                _registered_method=True)
        self.GetResourceLocation = channel.unary_unary(
                '/featureform.serving.proto.Feature/GetResourceLocation',
                request_serializer=featureform_dot_proto_dot_serving__pb2.ResourceIdRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.ResourceLocation.FromString,
                _registered_method=True)
        self.ServeFeatureView = channel.unary_unary(
                '/featureform.serving.proto.Feature/ServeFeatureView',
                request_serializer=featureform_dot_proto_dot_serving__pb2.FeatureViewServeRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.FeatureViewRows.FromString,
                _registered_method=True)


class FeatureServicer(object):
    """Missing associated documentation comment in .proto file."""

    def TrainingData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TrainTestSplit(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TrainingDataColumns(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FeatureServe(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SourceData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SourceColumns(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Nearest(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BatchFeatureServe(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetResourceLocation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServeFeatureView(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_FeatureServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TrainingData': grpc.unary_stream_rpc_method_handler(
                    servicer.TrainingData,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.TrainingDataRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.TrainingDataRows.SerializeToString,
            ),
            'TrainTestSplit': grpc.stream_stream_rpc_method_handler(
                    servicer.TrainTestSplit,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.TrainTestSplitRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.BatchTrainTestSplitResponse.SerializeToString,
            ),
            'TrainingDataColumns': grpc.unary_unary_rpc_method_handler(
                    servicer.TrainingDataColumns,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.TrainingDataColumnsRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.TrainingColumns.SerializeToString,
            ),
            'FeatureServe': grpc.unary_unary_rpc_method_handler(
                    servicer.FeatureServe,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.FeatureServeRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.FeatureRow.SerializeToString,
            ),
            'SourceData': grpc.unary_stream_rpc_method_handler(
                    servicer.SourceData,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.SourceDataRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.SourceDataRows.SerializeToString,
            ),
            'SourceColumns': grpc.unary_unary_rpc_method_handler(
                    servicer.SourceColumns,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.SourceColumnRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.SourceDataColumns.SerializeToString,
            ),
            'Nearest': grpc.unary_unary_rpc_method_handler(
                    servicer.Nearest,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.NearestRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.NearestResponse.SerializeToString,
            ),
            'BatchFeatureServe': grpc.unary_stream_rpc_method_handler(
                    servicer.BatchFeatureServe,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.BatchFeatureServeRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.BatchFeatureRows.SerializeToString,
            ),
            'GetResourceLocation': grpc.unary_unary_rpc_method_handler(
                    servicer.GetResourceLocation,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.ResourceIdRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.ResourceLocation.SerializeToString,
            ),
            'ServeFeatureView': grpc.unary_unary_rpc_method_handler(
                    servicer.ServeFeatureView,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.FeatureViewServeRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.FeatureViewRows.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'featureform.serving.proto.Feature', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('featureform.serving.proto.Feature', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Feature(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def TrainingData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/featureform.serving.proto.Feature/TrainingData',
            featureform_dot_proto_dot_serving__pb2.TrainingDataRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.TrainingDataRows.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TrainTestSplit(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/featureform.serving.proto.Feature/TrainTestSplit',
            featureform_dot_proto_dot_serving__pb2.TrainTestSplitRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.BatchTrainTestSplitResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TrainingDataColumns(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/featureform.serving.proto.Feature/TrainingDataColumns',
            featureform_dot_proto_dot_serving__pb2.TrainingDataColumnsRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.TrainingColumns.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def FeatureServe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/featureform.serving.proto.Feature/FeatureServe',
            featureform_dot_proto_dot_serving__pb2.FeatureServeRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.FeatureRow.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SourceData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/featureform.serving.proto.Feature/SourceData',
            featureform_dot_proto_dot_serving__pb2.SourceDataRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.SourceDataRows.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SourceColumns(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/featureform.serving.proto.Feature/SourceColumns',
            featureform_dot_proto_dot_serving__pb2.SourceColumnRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.SourceDataColumns.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Nearest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/featureform.serving.proto.Feature/Nearest',
            featureform_dot_proto_dot_serving__pb2.NearestRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.NearestResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def BatchFeatureServe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/featureform.serving.proto.Feature/BatchFeatureServe',
            featureform_dot_proto_dot_serving__pb2.BatchFeatureServeRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.BatchFeatureRows.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetResourceLocation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/featureform.serving.proto.Feature/GetResourceLocation',
            featureform_dot_proto_dot_serving__pb2.ResourceIdRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.ResourceLocation.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ServeFeatureView(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/featureform.serving.proto.Feature/ServeFeatureView',
            featureform_dot_proto_dot_serving__pb2.FeatureViewServeRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.FeatureViewRows.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Feature resource classes for Featureform.

This module contains classes for defining and managing features.
"""

from dataclasses import dataclass, field
from datetime import timedelta
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union

import dill

if TYPE_CHECKING:
    from ..registrar.realtime_feature import FeatureInput, RealtimeInput

from typeguard import typechecked

# Type alias for realtime feature inputs
RealtimeFeatureInput = Union["FeatureInput", "RealtimeInput"]

from ..config import ResourceSnowflakeConfig
from ..core import ResourceVariant
from ..enums import (
    AggregateFunction,
    ComputationMode,
    OperationType,
    ResourceStatus,
    ResourceType,
    ScalarType,
)
from ..operations.equivalence import _get_and_set_equivalent_variant
from ..proto import metadata_pb2 as pb
from ..types import VectorType, type_from_proto
from .base import NameVariant
from .mappings import ColumnSchema, ResourceColumnMapping
from .provider import Properties, ServerStatus
from .schedule import Schedule

# Type alias for resource location
ResourceLocation = ResourceColumnMapping


def _duration_to_timedelta(duration_proto: Optional["Duration"]) -> timedelta:
    """Convert protobuf Duration to Python timedelta."""
    from google.protobuf.duration_pb2 import Duration

    if duration_proto is None:
        return timedelta(0)

    duration = Duration()
    duration.CopyFrom(duration_proto)
    return duration.ToTimedelta()


@dataclass(frozen=True)
class AttributeFeature:
    """Represents a simple attribute feature.

    Attributes:
        name: Name assigned to the feature.
        input_column: Column used for the value of the feature.
        input_type: Type of the feature.
    """

    name: str
    input_column: str
    input_type: ScalarType

    def to_proto(self):
        return pb.AttributeFeatureDefinition(
            name=self.name,
            input_column=ColumnSchema(
                name=self.input_column,
                value_type=self.input_type,
            ).to_proto(),
        )

    @staticmethod
    def from_proto(attribute: pb.AttributeFeatureDefinition) -> "AttributeFeature":
        column = ColumnSchema.from_proto(attribute.input_column)

        return AttributeFeature(
            name=attribute.name,
            input_column=column.name,
            input_type=column.value_type,
        )


@dataclass(frozen=True)
class AggregateFeature:
    """Represents a time windowed aggregate feature."""

    name: str
    input_column: str
    input_type: ScalarType
    function: Union[AggregateFunction, str]
    time_window: timedelta

    def __post_init__(self) -> None:
        try:
            aggregate_function = AggregateFunction(self.function)
        except ValueError as error:
            allowed = ", ".join(AggregateFunction._valid_functions())
            raise ValueError(
                f"Unsupported aggregate function: {self.function}. Allowed values are: {allowed}"
            ) from error
        object.__setattr__(self, "function", aggregate_function)

    def to_proto(self):
        return pb.AggregateFeatureDefinition(
            name=self.name,
            input_column=ColumnSchema(
                name=self.input_column,
                value_type=self.input_type,
            ).to_proto(),
            function=AggregateFunction._aggregate_function_to_proto(
                self.function,
            ),
            time_window=self.time_window,
        )

    @staticmethod
    def from_proto(aggregate: pb.AggregateFeatureDefinition) -> "AggregateFeature":
        column = ColumnSchema.from_proto(aggregate.input_column)

        return AggregateFeature(
            name=aggregate.name,
            input_column=column.name,
            input_type=column.value_type,
            function=AggregateFunction._aggregate_function_from_proto(
                aggregate.function
            ),
            time_window=_duration_to_timedelta(aggregate.time_window),
        )


@dataclass
class RealtimeFeatureDefinition:
    """
    Definition for a realtime feature that executes a Python function at serving time.

    Realtime features combine real-time request data with pre-computed feature values
    to compute feature values on-the-fly during inference.
    """

    serialized_function: bytes
    function_source: str
    inputs: List[RealtimeFeatureInput]
    return_type: str
    requirements: List[str] = field(default_factory=list)

    def to_proto(self) -> pb.RealtimeFeatureDefinition:
        """Convert to protobuf message."""
        # Import here to avoid circular imports
        from ..registrar.realtime_feature import FeatureInput, RealtimeInput

        inputs_proto = []
        for inp in self.inputs:
            if isinstance(inp, FeatureInput):
                feature_input_proto = pb.FeatureInput(
                    feature=pb.NameVariant(
                        name=inp.feature_name,
                        variant=inp.feature_variant,
                    ),
                )
                inputs_proto.append(
                    pb.RealtimeFeatureInput(feature_input=feature_input_proto)
                )
            elif isinstance(inp, RealtimeInput):
                realtime_input_proto = pb.RealtimeInput(
                    parameter_name=inp.parameter_name,
                )
                # Add training feature if specified
                if inp.training_feature is not None:
                    realtime_input_proto.training_feature.CopyFrom(
                        pb.NameVariant(
                            name=inp.training_feature_name,
                            variant=inp.training_feature_variant,
                        )
                    )
                inputs_proto.append(
                    pb.RealtimeFeatureInput(realtime_input=realtime_input_proto)
                )

        return pb.RealtimeFeatureDefinition(
            serialized_function=self.serialized_function,
            function_source=self.function_source,
            inputs=inputs_proto,
            requirements=self.requirements,
            return_type=self.return_type,
        )

    @staticmethod
    def from_proto(
        proto: pb.RealtimeFeatureDefinition,
    ) -> "RealtimeFeatureDefinition":
        """Create from protobuf message."""
        from ..registrar.realtime_feature import FeatureInput, RealtimeInput

        inputs = []
        for inp in proto.inputs:
            if inp.HasField("feature_input"):
                fi = inp.feature_input
                inputs.append(
                    FeatureInput(
                        feature=fi.feature.name,
                        variant=fi.feature.variant,
                    )
                )
            elif inp.HasField("realtime_input"):
                ri = inp.realtime_input
                training_feature = None
                if ri.HasField("training_feature"):
                    training_feature = (
                        ri.training_feature.name,
                        ri.training_feature.variant,
                    )
                inputs.append(
                    RealtimeInput(
                        name=ri.parameter_name,
                        training_feature=training_feature,
                    )
                )

        return RealtimeFeatureDefinition(
            serialized_function=proto.serialized_function,
            function_source=proto.function_source,
            inputs=inputs,
            return_type=proto.return_type,
            requirements=list(proto.requirements),
        )


# Type alias for feature definition payloads
FeatureDefinitionPayload = Union[
    AttributeFeature, AggregateFeature, RealtimeFeatureDefinition
]


class FeatureDefinition:
    """Helper class for converting feature definitions to/from protobuf."""

    @staticmethod
    def to_proto(
        definition: Optional[FeatureDefinitionPayload],
    ) -> Optional[pb.FeatureDefinition]:
        if definition is None:
            return None

        if isinstance(definition, AttributeFeature):
            return pb.FeatureDefinition(
                attribute=definition.to_proto(),
            )

        if isinstance(definition, AggregateFeature):
            return pb.FeatureDefinition(
                aggregate=definition.to_proto(),
            )

        if isinstance(definition, RealtimeFeatureDefinition):
            return pb.FeatureDefinition(
                realtime=definition.to_proto(),
            )

        raise TypeError(f"Unsupported feature definition type: {type(definition)}")

    @staticmethod
    def from_proto(
        proto_definition: Optional[pb.FeatureDefinition],
    ) -> FeatureDefinitionPayload:
        if proto_definition is None:
            return None

        if proto_definition.HasField("attribute"):
            attribute = proto_definition.attribute
            return AttributeFeature.from_proto(attribute)

        if proto_definition.HasField("aggregate"):
            aggregate = proto_definition.aggregate
            return AggregateFeature.from_proto(aggregate)

        if proto_definition.HasField("realtime"):
            realtime = proto_definition.realtime
            return RealtimeFeatureDefinition.from_proto(realtime)

        return None


@typechecked
@dataclass
class Feature:
    """High-level feature resource containing multiple variants."""

    name: str
    default_variant: str
    variants: List[str]

    def to_dictionary(self):
        return {
            "name": self.name,
            "default_variant": self.default_variant,
            "variants": self.variants,
        }


class PrecomputedFeatureParameters:
    """Parameters for precomputed features."""

    pass


@typechecked
@dataclass
class OndemandFeatureParameters:
    """Parameters for on-demand features."""

    definition: str = ""

    def proto(self) -> pb.FeatureParameters:
        ondemand_feature_parameters = pb.OndemandFeatureParameters(
            definition=self.definition
        )
        feature_parameters = pb.FeatureParameters()
        feature_parameters.ondemand.CopyFrom(ondemand_feature_parameters)
        return feature_parameters


# Type alias for additional feature parameters
Additional_Parameters = Union[
    PrecomputedFeatureParameters, OndemandFeatureParameters, None
]


# FeatureVariant class
@typechecked
@dataclass
class FeatureVariant(ResourceVariant):
    name: str
    owner: str
    variant: str
    source: Optional[Any] = None
    value_type: Optional[Union[VectorType, ScalarType, str]] = None
    entity: str = ""
    location: Optional[ResourceLocation] = None
    description: str = ""
    provider: Optional[str] = None
    created: str = None
    tags: Optional[list] = None
    properties: Optional[dict] = None
    schedule: str = ""
    schedule_obj: Schedule = None
    status: str = "NO_STATUS"
    error: Optional[str] = None
    additional_parameters: Optional[Additional_Parameters] = None  # TODO Remove
    server_status: Optional[ServerStatus] = None
    resource_snowflake_config: Optional[ResourceSnowflakeConfig] = None
    feature_definition: Optional[FeatureDefinitionPayload] = None

    def __post_init__(self):
        if isinstance(self.value_type, str):
            self.value_type = ScalarType(self.value_type)

    @property
    def is_realtime(self) -> bool:
        """Check if this is a realtime feature."""
        return isinstance(self.feature_definition, RealtimeFeatureDefinition)

    def update_schedule(self, schedule) -> None:
        self.schedule_obj = Schedule(
            name=self.name,
            variant=self.variant,
            resource_type=4,
            schedule_string=schedule,
        )
        self.schedule = schedule

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.FEATURE_VARIANT

    def name_variant(self) -> NameVariant:
        return (self.name, self.variant)

    def get(self, stub) -> "FeatureVariant":
        return FeatureVariant.get_by_name_variant(stub, self.name, self.variant)

    @staticmethod
    def get_by_name_variant(stub, name, variant):
        name_variant = pb.NameVariantRequest(
            name_variant=pb.NameVariant(name=name, variant=variant)
        )
        feature = next(stub.GetFeatureVariants(iter([name_variant])))

        return FeatureVariant(
            created=None,
            name=feature.name,
            variant=feature.variant,
            source=(feature.source.name, feature.source.variant),
            value_type=type_from_proto(feature.type),
            entity=feature.entity,
            owner=feature.owner,
            provider=feature.provider,
            location=ResourceColumnMapping("", "", ""),
            description=feature.description,
            tags=list(feature.tags.tag),
            properties=dict(feature.properties.property.items()),
            status=feature.status.Status._enum_type.values[feature.status.status].name,
            error=feature.status.error_message,
            server_status=ServerStatus.from_proto(feature.status),
            additional_parameters=None,
            feature_definition=FeatureDefinition.from_proto(feature.feature_definition),
        )

    def _get_and_set_equivalent_variant(self, req_id, stub):
        if hasattr(self.source, "name_variant"):
            self.source = self.source.name_variant()

        feature_definition_proto = FeatureDefinition.to_proto(self.feature_definition)

        # Determine mode and status based on feature type
        if self.is_realtime:
            mode = ComputationMode.CLIENT_COMPUTED.proto()
            status = pb.ResourceStatus(status=pb.ResourceStatus.READY)
        else:
            mode = ComputationMode.PRECOMPUTED.proto()
            status = pb.ResourceStatus(status=pb.ResourceStatus.NO_STATUS)

        feature_variant_message = pb.FeatureVariant(
            name=self.name,
            variant=self.variant,
            source=(
                pb.NameVariant(name=self.source[0], variant=self.source[1])
                if self.source
                else None
            ),
            type=self.value_type.to_proto() if self.value_type else None,
            entity=self.entity,
            owner=self.owner,
            description=self.description,
            schedule=self.schedule,
            provider=self.provider,
            offline_store_provider=self.provider or "",
            columns=self.location.proto() if self.location else None,
            mode=mode,
            tags=pb.Tags(tag=self.tags or []),
            properties=Properties(self.properties).serialized,
            status=status,
            additional_parameters=None,
            resource_snowflake_config=(
                self.resource_snowflake_config.to_proto()
                if self.resource_snowflake_config
                else None
            ),
            feature_definition=feature_definition_proto,
        )

        # Initialize the FeatureVariantRequest message with the FeatureVariant message
        serialized = pb.FeatureVariantRequest(
            feature_variant=feature_variant_message,
            request_id="",
        )

        return (
            serialized,
            _get_and_set_equivalent_variant(
                req_id, serialized, "feature_variant", stub
            ),
            "feature_variant",
        )

    def _create(self, req_id, stub) -> Tuple[Optional[str], Optional[str]]:
        serialized, existing_variant, _ = self._get_and_set_equivalent_variant(
            req_id, stub
        )
        if existing_variant is None:
            stub.CreateFeatureVariant(serialized)
        return serialized.feature_variant.variant, existing_variant

    def get_status(self):
        return ResourceStatus(self.status)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value


# OnDemandFeatureVariant class
@typechecked
@dataclass
class OnDemandFeatureVariant(ResourceVariant):
    owner: str
    variant: str
    tags: List[str] = field(default_factory=list)
    properties: dict = field(default_factory=dict)
    name: str = ""
    description: str = ""
    status: str = "READY"
    error: Optional[str] = None
    additional_parameters: Optional[Additional_Parameters] = None
    server_status: Optional[ServerStatus] = None

    def __call__(self, fn):
        if self.description == "" and fn.__doc__ is not None:
            self.description = fn.__doc__
        if self.name == "":
            self.name = fn.__name__

        self.query = dill.dumps(fn.__code__)
        feature_text = dill.source.getsource(fn)
        self.additional_parameters = OndemandFeatureParameters(definition=feature_text)
        fn.name_variant = self.name_variant
        fn.query = self.query
        return fn

    def name_variant(self):
        return (self.name, self.variant)

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.ONDEMAND_FEATURE

    def _get_and_set_equivalent_variant(self, req_id, stub):
        serialized = pb.FeatureVariantRequest(
            feature_variant=pb.FeatureVariant(
                name=self.name,
                variant=self.variant,
                owner=self.owner,
                description=self.description,
                function=pb.PythonFunction(query=self.query),
                mode=ComputationMode.CLIENT_COMPUTED.proto(),
                tags=pb.Tags(tag=self.tags),
                properties=Properties(self.properties).serialized,
                status=pb.ResourceStatus(status=pb.ResourceStatus.READY),
                additional_parameters=self.additional_parameters.proto(),
            ),
            request_id="",
        )

        return (
            serialized,
            _get_and_set_equivalent_variant(
                req_id, serialized, "feature_variant", stub
            ),
            "feature_variant",
        )

    def _create(self, req_id, stub) -> Tuple[Optional[str], Optional[str]]:
        serialized, existing_variant, _ = self._get_and_set_equivalent_variant(
            req_id, stub
        )
        if existing_variant is None:
            stub.CreateFeatureVariant(serialized)
        return serialized.feature_variant.variant, existing_variant

    def get(self, stub) -> "OnDemandFeatureVariant":
        name_variant = pb.NameVariantRequest(
            name_variant=pb.NameVariant(name=self.name, variant=self.variant)
        )
        ondemand_feature = next(stub.GetFeatureVariants(iter([name_variant])))
        additional_Parameters = self._get_additional_parameters(ondemand_feature)

        return OnDemandFeatureVariant(
            name=ondemand_feature.name,
            variant=ondemand_feature.variant,
            owner=ondemand_feature.owner,
            description=ondemand_feature.description,
            tags=list(ondemand_feature.tags.tag),
            properties=dict(ondemand_feature.properties.property.items()),
            status=ondemand_feature.status.Status._enum_type.values[
                ondemand_feature.status.status
            ].name,
            error=ondemand_feature.status.error_message,
            additional_parameters=additional_Parameters,
        )

    def _get_additional_parameters(self, feature):
        return OndemandFeatureParameters(definition="() => FUNCTION")

    def get_status(self):
        return ResourceStatus(self.status)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Cleanup operations for Featureform resources.

This module contains classes and functions for cleaning up resources.
"""

from dataclasses import dataclass
from typing import List, Set, Tuple

from ..enums import ResourceType


@dataclass
class CleanupResult:
    """
    Result of a cleanup operation.

    Attributes:
        deleted_resources: Set of successfully deleted/queued resources as tuples (resource_type, name, variant)
        failed_resources: List of resources that failed to delete as tuples (resource_type, name, variant)
        error_count: Total number of errors encountered
        errors: List of tuples containing (resource_id, error_message)
    """

    deleted_resources: Set[Tuple[ResourceType, str, str]]
    failed_resources: List[Tuple[ResourceType, str, str]]
    error_count: int
    errors: List[Tuple[str, str]]

    @staticmethod
    def empty():
        return CleanupResult(set(), [], 0, [])

    @property
    def success_count(self) -> int:
        """Number of successfully deleted resources."""
        return len(self.deleted_resources)

    @property
    def total_attempted(self) -> int:
        """Total number of resources attempted to delete."""
        return self.success_count + len(self.failed_resources)

    @property
    def has_failures(self) -> bool:
        """Whether any resources failed to delete."""
        return len(self.failed_resources) > 0

    @property
    def is_complete_success(self) -> bool:
        """Whether all resources were successfully deleted."""
        return not self.has_failures

    def get_failed_by_type(self) -> dict:
        """Group failed resources by type."""
        failed_by_type = {}
        for res_type, name, variant in self.failed_resources:
            if res_type not in failed_by_type:
                failed_by_type[res_type] = []
            failed_by_type[res_type].append((name, variant))
        return failed_by_type

    def get_deleted_by_type(self) -> dict:
        """Group deleted resources by type."""
        deleted_by_type = {}
        for res_type, name, variant in self.deleted_resources:
            if res_type not in deleted_by_type:
                deleted_by_type[res_type] = []
            deleted_by_type[res_type].append((name, variant))
        return deleted_by_type



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Protocol definitions for Featureform resources.

This module defines protocols (structural typing) for resources to enable
duck typing and reduce coupling.
"""

from typing import Protocol, Tuple, runtime_checkable

from ..enums import OperationType, ResourceType


@runtime_checkable
class HasNameVariant(Protocol):
    """Protocol for resources that have name and variant attributes."""

    def name_variant(self) -> Tuple[str, str]:
        """Returns the name and variant of the resource"""
        ...


@runtime_checkable
class ResourceProtocol(Protocol):
    """Protocol for all Featureform resources."""

    name: str

    def get_resource_type(self) -> ResourceType:
        """Return the resource type."""
        ...

    def operation_type(self) -> OperationType:
        """Return the operation type (CREATE or GET)."""
        ...


@runtime_checkable
class ResourceVariant(HasNameVariant, Protocol):
    """Protocol for resource variants with name, variant, and status."""

    name: str
    variant: str
    server_status: "ServerStatus"  # Forward reference

    def get_resource_type(self) -> ResourceType:
        """Returns the type of the resource"""
        ...

    def to_key(self) -> Tuple[ResourceType, str, str]:
        """Returns a tuple key of (resource_type, name, variant)"""
        return self.get_resource_type(), self.name, self.variant



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Configuration classes for file store providers.
"""

import json
from dataclasses import dataclass, field
from typing import List, Optional, Union

from typeguard import typechecked

from ..utils import is_valid_url
from .credentials import (
    AWSAssumeRoleCredentials,
    AWSStaticCredentials,
    BasicCredentials,
    GCPCredentials,
    KerberosCredentials,
)


@typechecked
@dataclass
class S3StoreConfig:
    bucket_path: str
    bucket_region: str
    credentials: Union[AWSStaticCredentials, AWSAssumeRoleCredentials]
    path: str = ""

    def __post_init__(self):
        # Validate that the bucket_path does not end with a slash
        if self.bucket_path.endswith("/"):
            raise ValueError("The 'bucket_path' cannot end with '/'.")

    def software(self) -> str:
        return "S3"

    def type(self) -> str:
        return "S3"

    def serialize(self) -> bytes:
        config = self.config()
        return bytes(json.dumps(config), "utf-8")

    def config(self):
        return {
            "Credentials": self.credentials.config(),
            "BucketRegion": self.bucket_region,
            "BucketPath": self.bucket_path,
            "Path": self.path,
        }

    def store_type(self):
        return self.type()


@typechecked
@dataclass
class GCSFileStoreConfig:
    credentials: GCPCredentials
    bucket_name: str
    bucket_path: str = ""

    def software(self) -> str:
        return "gcs"

    def type(self) -> str:
        return "GCS"

    def serialize(self) -> bytes:
        config = {
            "BucketName": self.bucket_name,
            "BucketPath": self.bucket_path,
            "Credentials": self.credentials.config(),
        }
        return bytes(json.dumps(config), "utf-8")

    def config(self):
        return {
            "BucketName": self.bucket_name,
            "BucketPath": self.bucket_path,
            "Credentials": self.credentials.config(),
        }

    def store_type(self):
        return self.type()


@typechecked
@dataclass
class AzureFileStoreConfig:
    account_name: str
    account_key: str
    container_name: str
    root_path: str

    def software(self) -> str:
        return "azure"

    def type(self) -> str:
        return "AZURE"

    def serialize(self) -> bytes:
        config = {
            "AccountName": self.account_name,
            "AccountKey": self.account_key,
            "ContainerName": self.container_name,
            "Path": self.root_path,
        }
        return bytes(json.dumps(config), "utf-8")

    def config(self):
        return {
            "AccountName": self.account_name,
            "AccountKey": self.account_key,
            "ContainerName": self.container_name,
            "Path": self.root_path,
        }

    def store_type(self):
        return self.type()


@typechecked
@dataclass
class VaultConfig:
    address: str
    token: str
    mount_path: str = "secret"
    default_secret_path: str = ""

    def __post_init__(self):
        if not self.address:
            raise ValueError("The 'address' field cannot be empty.")
        if not self.token:
            raise ValueError("The 'token' field cannot be empty.")
        if not self.mount_path:
            raise ValueError("The 'mount_path' field cannot be empty.")

        if not is_valid_url(self.address):
            raise ValueError(
                "The 'address' field must be in the form of schema://host:port"
            )

    def software(self) -> str:
        return "vault"

    def type(self) -> str:
        return "VAULT"

    def config(self):
        return {
            "Address": self.address,
            "Token": self.token,
            "MountPath": self.mount_path,
            "DefaultSecretPath": self.default_secret_path,
        }

    def serialize(self) -> bytes:
        return bytes(json.dumps(self.config()), "utf-8")

    @classmethod
    def deserialize(cls, json_bytes: bytes) -> "VaultConfig":
        config = json.loads(json_bytes)
        return cls(
            address=config["Address"],
            token=config["Token"],
            mount_path=config["MountPath"],
            default_secret_path=config["DefaultSecretPath"],
        )


@typechecked
@dataclass
class KafkaConfig:
    bootstrap_servers: List[str]
    use_msk_iam_auth: bool = False
    options: Optional[dict] = field(default_factory=dict)

    def software(self) -> str:
        return "kafka"

    def type(self) -> str:
        return "KAFKA"

    def config(self):
        return {
            "BootstrapServers": self.bootstrap_servers,
            "UseMSKIAMAuth": self.use_msk_iam_auth,
            "Options": self.options,
        }

    def serialize(self) -> bytes:
        return bytes(json.dumps(self.config()), "utf-8")

    @classmethod
    def deserialize(cls, json_str: bytes) -> "KafkaConfig":
        config = json.loads(json_str)
        return cls(
            bootstrap_servers=config["BootstrapServers"],
            use_msk_iam_auth=config["UseMSKIAMAuth"],
            options=config.get("Options"),
        )


def read_file(file_path: str) -> str:
    """Read file content as string."""
    with open(file_path, "r") as f:
        return f.read()


@typechecked
@dataclass
class HDFSConfig:
    def __init__(
        self,
        host: str,
        port: str,
        path: str,
        hdfs_site_file: str = "",
        core_site_file: str = "",
        hdfs_site_contents: str = "",
        core_site_contents: str = "",
        credentials: Union[BasicCredentials, KerberosCredentials] = BasicCredentials(
            ""
        ),
    ):
        bucket_path_ends_with_slash = len(path) != 0 and path[-1] == "/"

        if bucket_path_ends_with_slash:
            raise Exception("The 'bucket_path' cannot end with '/'.")

        self.path = path
        self.host = host
        self.port = port
        self.credentials = credentials

        self.hdfs_site_conf = self.__get_hdfs_site_contents(
            hdfs_site_file, hdfs_site_contents
        )
        self.core_site_conf = self.__get_core_site_contents(
            core_site_file, core_site_contents
        )

    def __get_hdfs_site_contents(self, site_file, site_contents):
        return self.__get_site_contents(site_file, site_contents, "hdfs")

    def __get_core_site_contents(self, site_file, site_contents):
        return self.__get_site_contents(site_file, site_contents, "core")

    @staticmethod
    def __get_site_contents(site_file, site_contents, config_type):
        if site_file == "" and site_contents == "":
            raise ValueError(
                f"{config_type} site config must be provided. Either provide a path to the xml file "
                f"({config_type}_site_file) or the contents of the file ({config_type}_site_contents)."
            )

        elif site_file != "" and site_contents != "":
            raise ValueError(
                f"Only one of {config_type}_site_file or {config_type}_site_contents should be provided."
            )

        elif site_file != "" and site_contents == "":
            return read_file(site_file)

        else:
            return site_contents

    def software(self) -> str:
        return "HDFS"

    def type(self) -> str:
        return "HDFS"

    def serialize(self) -> bytes:
        config = self.config()
        return bytes(json.dumps(config), "utf-8")

    def config(self):
        return {
            "Host": self.host,
            "Port": self.port,
            "Path": self.path,
            "HDFSSiteConf": self.hdfs_site_conf,
            "CoreSiteConf": self.core_site_conf,
            "CredentialType": self.credentials.type(),
            "CredentialConfig": self.credentials.config(),
        }

    def store_type(self):
        return self.type()


@typechecked
@dataclass
class OnlineBlobConfig:
    store_type: str
    store_config: dict

    def software(self) -> str:
        return self.store_type

    def type(self) -> str:
        return "BLOB_ONLINE"

    def config(self):
        return self.store_config

    def serialize(self) -> bytes:
        config = {
            "Type": self.store_type,
            "Config": self.store_config,
        }
        return bytes(json.dumps(config), "utf-8")

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, OnlineBlobConfig):
            return False
        return (
            self.store_type == __value.store_type
            and self.store_config == __value.store_config
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Base classes and types for Featureform resources.

This module contains base classes, protocols, and the Resource union type that are used
across all resource modules.
"""

from abc import ABC
from typing import Tuple, Union

from typeguard import typechecked

# Type alias for name-variant tuples
NameVariant = Tuple[str, str]


@typechecked
def valid_name_variant(nvar: NameVariant) -> bool:
    """Check if a name-variant tuple is valid (both parts non-empty)."""
    return nvar[0] != "" and nvar[1] != ""


class Location(ABC):
    """Abstract base class for resources that have a location."""

    def resource_identifier(self):
        """
        Return the location of the resource.
        :return:
        """
        raise NotImplementedError


# Resource union type - defined here to avoid circular imports
# This needs to be defined after all resource classes are imported
Resource = Union[
    "PrimaryData",
    "Provider",
    "Entity",
    "User",
    "FeatureVariant",
    "LabelVariant",
    "TrainingSetVariant",
    "SourceVariant",
    "Schedule",
    "ProviderReference",
    "SourceReference",
    "EntityReference",
    "Model",
    "OnDemandFeatureVariant",
    "FeatureView",
    "KafkaTopic",
]


# Config union type - includes all provider configuration types
# Import at runtime to avoid circular imports
def _get_config_union():
    """Lazy import to avoid circular dependencies."""
    from ..config import (
        AzureFileStoreConfig,
        BigQueryConfig,
        CassandraConfig,
        ClickHouseConfig,
        DynamodbConfig,
        EmptyConfig,
        FirestoreConfig,
        GCSFileStoreConfig,
        HDFSConfig,
        K8sConfig,
        KafkaConfig,
        MongoDBConfig,
        OnlineBlobConfig,
        PineconeConfig,
        PostgresConfig,
        RedisConfig,
        RedshiftConfig,
        S3StoreConfig,
        SnowflakeConfig,
        SparkConfig,
        VaultConfig,
        WeaviateConfig,
    )

    return Union[
        RedisConfig,
        PineconeConfig,
        SnowflakeConfig,
        PostgresConfig,
        ClickHouseConfig,
        RedshiftConfig,
        BigQueryConfig,
        FirestoreConfig,
        SparkConfig,
        OnlineBlobConfig,
        AzureFileStoreConfig,
        S3StoreConfig,
        K8sConfig,
        MongoDBConfig,
        GCSFileStoreConfig,
        EmptyConfig,
        HDFSConfig,
        WeaviateConfig,
        DynamodbConfig,
        CassandraConfig,
        KafkaConfig,
        VaultConfig,
    ]


# Create the Config type at module level
Config = _get_config_union()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Transformation classes for Featureform resources.

This module contains all transformation-related classes including Transformation base class,
SQLTransformation, DFTransformation, and PrimaryData.
"""

import re
from abc import ABC
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from typeguard import typechecked

from ..config.compute import EmptySparkFlags, K8sArgs, SparkFlags
from ..config.offline_stores import ResourceSnowflakeConfig
from ..enums import LocationType, SourceType
from ..proto import metadata_pb2 as pb
from .schedule import PartitionType

if TYPE_CHECKING:
    pass


class Transformation(ABC):
    """Base class for all transformations."""

    @classmethod
    def from_proto(cls, source_transformation: pb.Transformation):
        # Import here to avoid circular dependency
        transformation_type = source_transformation.WhichOneof("type")
        if transformation_type == "DFTransformation":
            return DFTransformation.from_proto(source_transformation.DFTransformation)
        elif transformation_type == "SQLTransformation":
            return SQLTransformation.from_proto(source_transformation.SQLTransformation)
        else:
            raise Exception(f"Invalid transformation type {source_transformation}")


@typechecked
@dataclass
class SQLTransformation(Transformation):
    """SQL-based transformation with query and optional parameters."""

    query: str
    args: Optional[K8sArgs] = None
    func_params_to_inputs: Dict[str, Any] = field(default_factory=dict)
    is_incremental: bool = False
    incremental_inputs: list = field(default_factory=list)
    partition_options: Optional[PartitionType] = None
    spark_flags: SparkFlags = field(default_factory=lambda: EmptySparkFlags)
    resource_snowflake_config: Optional[ResourceSnowflakeConfig] = None

    _sql_placeholder_regex: str = field(
        default=r"\{\{\s*\w+\s*\}\}", init=False, repr=False
    )

    def __post_init__(self):
        self._validate_inputs_to_func_params(self.func_params_to_inputs)

    def type(self) -> str:
        """Return the type of the SQL transformation."""
        return SourceType.SQL_TRANSFORMATION.value

    def kwargs(self) -> Dict[str, pb.Transformation]:
        return {"transformation": self.to_proto()}

    @classmethod
    def from_proto(cls, sql_transformation: pb.SQLTransformation):
        return SQLTransformation(sql_transformation.query)

    def _validate_inputs_to_func_params(self, inputs: Dict[str, Any]) -> None:
        # Find and replace placeholders in the query with source name variants
        for placeholder in self._get_placeholders():
            clean_placeholder = placeholder.strip(" {}")
            if clean_placeholder not in self.func_params_to_inputs.keys():
                raise ValueError(
                    f"SQL placeholder '{placeholder}' not found in input arguments. "
                    f"Available input arguments: {', '.join(inputs.keys())}.\n"
                    f"Expected inputs based on function parameters: {', '.join(self.func_params_to_inputs.keys())}."
                )

    def _resolve_input_variants(self) -> Dict[str, Any]:
        """Resolve inputs to their name variants."""
        return {
            func_param: inp.name_variant() if hasattr(inp, "name_variant") else inp
            for func_param, inp in self.func_params_to_inputs.items()
        }

    def _get_placeholders(self) -> List[str]:
        """Get placeholders from the query."""
        return re.findall(self._sql_placeholder_regex, self.query)

    def to_proto(self) -> pb.Transformation:
        input_to_name_variant = self._resolve_input_variants()

        for i, inp in enumerate(self.incremental_inputs):
            if hasattr(inp, "name_variant"):
                self.incremental_inputs[i] = inp.name_variant()

        incremental_nvs = []
        for inp in self.incremental_inputs:
            incremental_nvs.append(pb.NameVariant(name=inp[0], variant=inp[1]))

        # Find and replace placeholders in the query with source name variants
        final_query = self.query
        for placeholder in self._get_placeholders():
            clean_placeholder = placeholder.strip(" {}")
            name_variant = input_to_name_variant[clean_placeholder]
            replacement = "{{ " + f"{name_variant[0]}.{name_variant[1]}" + " }}"
            final_query = final_query.replace(placeholder, replacement)

        partition_kwargs = {}
        if self.partition_options is not None:
            partition_kwargs = self.partition_options.proto_kwargs()

        # Construct the SQLTransformation protobuf message
        transformation = pb.Transformation(
            SQLTransformation=pb.SQLTransformation(
                query=final_query,
                is_incremental=self.is_incremental,
                incremental_source=incremental_nvs,
                resource_snowflake_config=(
                    self.resource_snowflake_config.to_proto()
                    if self.resource_snowflake_config
                    else None
                ),
            ),
            spark_flags=self.spark_flags.to_proto() if self.spark_flags else None,
            **partition_kwargs,
        )

        # Apply args transformations if any
        if self.args is not None:
            transformation = self.args.apply(transformation)

        return transformation


@typechecked
@dataclass
class DFTransformation(Transformation):
    """DataFrame transformation for processing data using Python/Spark."""

    query: bytes
    inputs: list
    args: K8sArgs = None
    source_text: str = ""
    canonical_func_text: str = ""
    is_incremental: bool = False
    incremental_inputs: list = field(default_factory=list)
    partition_options: Optional[PartitionType] = None
    spark_flags: SparkFlags = field(default_factory=lambda: EmptySparkFlags)

    def type(self):
        return SourceType.DF_TRANSFORMATION.value

    def to_proto(self) -> pb.Transformation:
        # Create a new list of name variants without modifying self.inputs
        name_variants = []
        for inp in self.inputs:
            if hasattr(inp, "name_variant"):
                name_variants.append(inp.name_variant())
            else:
                name_variants.append(inp)

        name_variant_protos = [
            pb.NameVariant(name=inp[0], variant=inp[1]) for inp in name_variants
        ]

        partition_kwargs = {}
        if self.partition_options is not None:
            partition_kwargs = self.partition_options.proto_kwargs()

        transformation = pb.Transformation(
            DFTransformation=pb.DFTransformation(
                query=self.query,
                inputs=name_variant_protos,
                source_text=self.source_text,
                is_incremental=self.is_incremental,
                incremental_sources=name_variant_protos,
                canonical_func_text=self.canonical_func_text,
            ),
            spark_flags=self.spark_flags.to_proto() if self.spark_flags else None,
            **partition_kwargs,
        )

        if self.args is not None:
            transformation = self.args.apply(transformation)

        return transformation

    def kwargs(self):
        return {"transformation": self.to_proto()}

    @classmethod
    def from_proto(cls, df_transformation: pb.DFTransformation):
        return DFTransformation(
            query=df_transformation.query,
            inputs=[(input.name, input.variant) for input in df_transformation.inputs],
            source_text=df_transformation.source_text,
        )


@dataclass
class PrimaryData:
    """Primary data source with location and optional timestamp column."""

    location: Any  # Location type - using Any to avoid circular import with typeguard
    timestamp_column: str = ""

    def kwargs(self) -> dict:
        # Import here to avoid circular dependency
        from .locations import (
            FileStore,
            GlueCatalogTable,
            SQLTable,
            StreamingInput,
            UnityCatalogTable,
        )

        primary_data_kwargs: Dict[str, Any] = {
            "timestamp_column": self.timestamp_column,
        }

        if isinstance(self.location, SQLTable):
            primary_data_kwargs["table"] = pb.SQLTable(
                schema=self.location.schema,
                database=self.location.database,
                name=self.location.name,
            )
        elif isinstance(self.location, FileStore):
            primary_data_kwargs["filestore"] = pb.FileStoreTable(
                path=self.location.resource_identifier(),
            )
        elif isinstance(self.location, GlueCatalogTable):
            primary_data_kwargs["catalog"] = pb.CatalogTable(
                catalog_type=pb.GLUE,
                database=self.location.database,
                table=self.location.table,
                table_format=self.location.table_format,
            )
        elif isinstance(self.location, UnityCatalogTable):
            primary_data_kwargs["catalog"] = pb.CatalogTable(
                catalog_type=pb.UNITY,
                database=self.location.database,
                schema=self.location.schema,
                table=self.location.table,
                table_format=self.location.table_format,
            )
        elif isinstance(self.location, StreamingInput):
            primary_data_kwargs["stream_channel"] = pb.StreamChannel(
                name=self.location.name,
                channel_name=self.location.topic,
                provider=self.location.provider,
            )
        else:
            raise ValueError(f"Unsupported location type: {type(self.location)}")

        return {"primaryData": pb.PrimaryData(**primary_data_kwargs)}

    def path(self) -> str:
        return self.location.resource_identifier()

    @staticmethod
    def from_proto(source_primary_data):
        # Import here to avoid circular dependency
        from .locations import FileStore, GlueCatalogTable, SQLTable, UnityCatalogTable

        primary_type = source_primary_data.WhichOneof("location")
        if primary_type == LocationType.TABLE:
            location = SQLTable.from_proto(source_primary_data.table)
        elif primary_type == LocationType.FILESTORE:
            location = FileStore.from_proto(source_primary_data.filestore)
        elif primary_type == LocationType.CATALOG:
            if source_primary_data.catalog.catalog_type == pb.GLUE:
                location = GlueCatalogTable.from_proto(source_primary_data.catalog)
            elif source_primary_data.catalog.catalog_type == pb.UNITY:
                location = UnityCatalogTable.from_proto(source_primary_data.catalog)
            else:
                raise Exception(
                    f"Unknown catalog type '{source_primary_data.catalog.catalog_type}'"
                )
        else:
            raise Exception(f"Invalid primary data type {source_primary_data}")

        return PrimaryData(location, source_primary_data.timestamp_column)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Online provider wrapper classes for Featureform.
"""

__all__ = ["OnlineProvider"]


class OnlineProvider:
    def __init__(self, registrar, provider):
        self.__registrar = registrar
        self.__provider = provider

    def name(self) -> str:
        return self.__provider.name

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, OnlineProvider):
            return False
        return (
            self.__provider == __value.__provider
            and self.__registrar == __value.__registrar
        )



"""
Feature API v2 - Declarative Feature Definitions.

This module provides the Feature API v2, which offers both a fluent builder pattern
for defining features and backward compatibility with the legacy API. It consolidates
all feature definition functionality into a single cohesive module.

Public API:
    - FeatureBuilder: The builder class for declarative feature definitions
    - Feature(): Factory function for legacy API compatibility
    - AggregateFunction: Enum of aggregation functions (re-exported from enums)

Supporting Classes:
    - AggregateConfig: Configuration for aggregations
    - BackfillConfig: Configuration for stream backfill
    - WindowedFeatureRef: Reference to a specific time window
    - FeatureSourceType: Enum for batch vs stream sources

Example Usage:
    ```python
    import featureform as ff
    from datetime import timedelta

    @ff.entity
    class User:
        # Builder Pattern - validated automatically during registration
        transaction_count = (
            ff.Feature()
            .from_dataset(
                batch_features,
                entity="CustomerID",
                values="Amount",
                timestamp="Timestamp",
            )
            .aggregate(
                function=ff.AggregateFunction.COUNT,
                windows=[timedelta(days=7)]
            )
        )

        # Legacy API (still supported)
        legacy_feature = ff.Feature(
            transformation[["CustomerID", "Amount", "Timestamp"]],
            type=ff.Float32,
            variant="v1",
        )
    ```
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

from ..enums import AggregateFunction, ResourceType, ScalarType
from ..resources import (
    AggregateFeature,
    AttributeFeature,
    FeatureVariant,
    ResourceColumnMapping,
)

__all__ = [
    # Main API
    "Feature",
    "FeatureBuilder",
    # Re-exported enum
    "AggregateFunction",
    # Supporting classes and enums
    "FeatureSourceType",
    "FeatureType",
    "AggregateConfig",
    "BackfillConfig",
    "BuiltFeatures",
    "AggregateBuiltFeatures",
    "AttributeBuiltFeature",
    # Utility functions
    "format_window_suffix",
]


# =============================================================================
# Supporting Classes
# =============================================================================


class FeatureSourceType(str, Enum):
    """Type of source for a feature."""

    BATCH = "batch"
    STREAM = "stream"


class FeatureType(str, Enum):
    """Type of feature.

    This enum distinguishes between different kinds of features,
    which determines how they can be accessed and used.
    """

    AGGREGATE = (
        "aggregate"  # Time-windowed aggregate features (e.g., COUNT over 7 days)
    )
    ATTRIBUTE = "attribute"  # Simple point-in-time attribute features
    REALTIME = "realtime"  # Realtime features computed at serving time


def format_window_suffix(window: timedelta) -> str:
    """Format a timedelta as a string suffix (e.g., '7d', '1h')."""
    total_seconds = int(window.total_seconds())
    if total_seconds >= 86400 and total_seconds % 86400 == 0:
        return f"{total_seconds // 86400}d"
    elif total_seconds >= 3600 and total_seconds % 3600 == 0:
        return f"{total_seconds // 3600}h"
    elif total_seconds >= 60 and total_seconds % 60 == 0:
        return f"{total_seconds // 60}m"
    else:
        return f"{total_seconds}s"


@dataclass
class SelectConfig:
    """Configuration for selected columns in a feature definition."""

    entity: str
    values: str
    timestamp: str = ""


@dataclass
class AggregateConfig:
    """Configuration for feature aggregation."""

    function: AggregateFunction
    windows: List[timedelta]


@dataclass
class BackfillConfig:
    """Configuration for backfilling stream features from batch data."""

    source: Any
    entity: str
    values: str
    timestamp: str


class BuiltFeatures(ABC):
    """
    Abstract base class for built features.

    This is the result of building a FeatureBuilder. Concrete implementations are:
    - AggregateBuiltFeatures: For features with time windows (e.g., COUNT over 7 days)
    - AttributeBuiltFeature: For simple attribute features without time windows

    Each FeatureVariant has its own variant field that can be updated during
    equivalence resolution.

    This wrapper is set as the entity attribute (e.g., User.transaction_count)
    and provides access to individual features.

    Example:
        # Aggregate features - use bracket indexing
        User.transaction_count[timedelta(days=7)]

        # Attribute features - use name_variant()
        User.account_age.name_variant()

        # Both types - get all features
        User.transaction_count.get_all_features()
    """

    def __init__(self, base_name: str, base_variant: str):
        self._base_name = base_name
        self._base_variant = base_variant

    @property
    @abstractmethod
    def feature_type(self) -> FeatureType:
        """Return the type of this built feature."""
        ...

    @abstractmethod
    def get_all_features(self) -> List[FeatureVariant]:
        """Return all FeatureVariant objects."""
        ...

    @abstractmethod
    def name_variant(self) -> Tuple[str, str]:
        """Return the (name, variant) tuple.

        For aggregate features, raises an error directing users to use bracket indexing.
        For attribute features, returns the feature's name and variant.
        """
        ...

    @abstractmethod
    def __getitem__(self, window: timedelta) -> FeatureVariant:
        """Get the FeatureVariant for a specific time window.

        For aggregate features, returns the FeatureVariant for the specified window.
        For attribute features, raises TypeError.
        """
        ...

    def get_resource_type(self) -> ResourceType:
        return ResourceType.FEATURE_VARIANT

    @abstractmethod
    def __repr__(self) -> str: ...


class AggregateBuiltFeatures(BuiltFeatures):
    """
    Built features for aggregate features with time windows.

    Contains multiple FeatureVariant objects, one per time window (e.g., 1d, 7d, 30d).
    Access specific windows using bracket notation: `feature[timedelta(days=7)]`

    Example:
        @ff.entity
        class User:
            transaction_count = (
                ff.Feature()
                .from_dataset(
                    transactions,
                    entity="user_id",
                    values="amount",
                    timestamp="ts",
                )
                .aggregate(
                    function=ff.AggregateFunction.COUNT,
                    windows=[timedelta(days=1), timedelta(days=7)]
                )
            )

        # Access specific window
        User.transaction_count[timedelta(days=7)]

        # Get all windows
        User.transaction_count.get_all_features()
    """

    def __init__(
        self,
        features: Dict[timedelta, FeatureVariant],
        base_name: str,
        base_variant: str,
    ):
        super().__init__(base_name, base_variant)
        self._features = features

    @property
    def feature_type(self) -> FeatureType:
        return FeatureType.AGGREGATE

    def __getitem__(self, window: timedelta) -> FeatureVariant:
        """Get the FeatureVariant for a specific time window.

        Args:
            window: The time window duration (e.g., timedelta(days=7))

        Returns:
            FeatureVariant for the specified window

        Raises:
            KeyError: If the specified window doesn't exist
        """
        if window not in self._features:
            available = [format_window_suffix(w) for w in self._features.keys()]
            raise KeyError(
                f"Window {format_window_suffix(window)} not found. "
                f"Available windows: {available}"
            )
        return self._features[window]

    def name_variant(self) -> Tuple[str, str]:
        """Raises ValueError - use bracket indexing for aggregate features."""
        raise ValueError(
            "Use bracket indexing (e.g., feature[timedelta(days=7)]) to access "
            "specific windows for aggregate features."
        )

    def get_all_features(self) -> List[FeatureVariant]:
        """Return all FeatureVariant objects for all time windows."""
        return list(self._features.values())

    def __repr__(self) -> str:
        windows = [format_window_suffix(w) for w in self._features.keys()]
        return f"AggregateBuiltFeatures({self._base_name}, windows={windows})"


class AttributeBuiltFeature(BuiltFeatures):
    """
    Built feature for an attribute feature without time windows.

    Contains a single FeatureVariant representing a simple point-in-time attribute.

    Example:
        @ff.entity
        class User:
            account_age = (
                ff.Feature()
                .from_dataset(
                    user_profiles,
                    entity="user_id",
                    values="age_days",
                    timestamp="updated_at",
                )
            )

        # Get name and variant
        User.account_age.name_variant()

        # Get the feature
        User.account_age.get_all_features()
    """

    def __init__(
        self,
        feature: FeatureVariant,
        base_name: str,
        base_variant: str,
    ):
        super().__init__(base_name, base_variant)
        self._feature = feature

    @property
    def feature_type(self) -> FeatureType:
        return FeatureType.ATTRIBUTE

    def __getitem__(self, window: timedelta) -> FeatureVariant:
        """Raises TypeError - attribute features cannot be indexed."""
        raise TypeError(
            f"'{self._base_name}' is not an aggregate feature and cannot be indexed. "
            f"Use name_variant() or get_all_features() instead."
        )

    def name_variant(self) -> Tuple[str, str]:
        """Return the (name, variant) tuple for this attribute feature."""
        return self._feature.name_variant()

    def get_all_features(self) -> List[FeatureVariant]:
        """Return the single FeatureVariant as a list."""
        return [self._feature]

    def __repr__(self) -> str:
        return f"AttributeBuiltFeature({self._base_name}, variant={self._base_variant})"


# =============================================================================
# FeatureBuilder Class
# =============================================================================


class FeatureBuilder:
    """
    Builder class for declaratively defining features using a fluent API.

    This class provides a builder pattern for defining features with method chaining.
    Features are automatically validated and registered when used with the @ff.entity
    decorator.

    Example:
        ```python
        @ff.entity
        class User:
            # Simple feature
            account_balance = (
                ff.Feature()
                .from_dataset(
                    batch_features,
                    entity="CustomerID",
                    values="Balance",
                    timestamp="Timestamp",
                )
            )

            # Aggregated feature with multiple time windows
            transaction_count = (
                ff.Feature()
                .from_dataset(
                    batch_features,
                    entity="CustomerID",
                    values="Amount",
                    timestamp="Timestamp",
                )
                .aggregate(
                    function=ff.AggregateFunction.COUNT,
                    windows=[timedelta(days=1), timedelta(days=7)]
                )
            )
        ```
    """

    def __init__(self):
        """Initialize an empty FeatureBuilder for the builder pattern."""
        # Feature identification
        # base_name: The base feature name (e.g., "transaction_count")
        #   For aggregate features, window suffixes are appended (e.g., "transaction_count_7d")
        self.base_name: Optional[str] = None

        # initial_variant: The user's proposed variant for all generated features.
        #
        # This field is named "initial" because the value represents a proposed variant
        # that may be changed during the registration flow. When `create_all()` is called,
        # the server's equivalence resolution checks if an existing feature with matching
        # semantics already exists. If an equivalent is found, the actual
        # `FeatureVariant.variant` will be updated to match the existing equivalent variant,
        # overriding the value specified here.
        #
        # Flow:
        #   1. User sets initial_variant via variant() builder method
        #   2. FeatureVariant objects are created with variant=initial_variant
        #   3. During create_all(), _get_and_set_equivalent_variant() queries the server
        #   4. If equivalent found: FeatureVariant.variant is updated to the existing variant
        #   5. If no equivalent: FeatureVariant.variant remains as initial_variant
        self.initial_variant: Optional[str] = None
        self.entity: Optional[Any] = None

        # Feature metadata
        self.owner: str = ""
        self.description: str = ""
        self.tags: List[str] = []
        self.properties: Dict[str, str] = {}
        self.value_type: Optional[Union[ScalarType, str]] = None

        # Source configuration
        self.source: Optional[Any] = None
        self.source_type: Optional[FeatureSourceType] = None

        # Column selection
        self.select_config: Optional[SelectConfig] = None

        # Aggregation configuration
        self.aggregate_config: Optional[AggregateConfig] = None

        # Backfill configuration (for stream features)
        self.backfill_config: Optional[BackfillConfig] = None

        # Built features (set after build() is called)
        self._built_features: Optional[BuiltFeatures] = None

    # =========================================================================
    # Builder Pattern Methods
    # =========================================================================

    def from_dataset(
        self,
        source: Any,
        entity: str,
        values: str,
        timestamp: str = "",
    ) -> "FeatureBuilder":
        """
        Set the batch dataset source for this feature and select columns.

        Args:
            source: A registered source (transformation result or table source)
            entity: Column name containing the entity identifier
            values: Column name containing the feature value
            timestamp: Column name containing the timestamp (optional)

        Returns:
            self: For method chaining

        Raises:
            ValueError: If a source has already been set
        """
        if self.source is not None:
            raise ValueError(
                "Source already set. from_dataset() or from_stream() can only be called once."
            )
        self.source = source
        self.source_type = FeatureSourceType.BATCH
        self.select_config = SelectConfig(
            entity=entity,
            values=values,
            timestamp=timestamp,
        )
        return self

    def from_stream(
        self,
        source: Any,
        entity: str,
        values: str,
        timestamp: str = "",
    ) -> "FeatureBuilder":
        """
        Set the streaming source for this feature and select columns.

        Args:
            source: A registered streaming source
            entity: Column name containing the entity identifier
            values: Column name containing the feature value
            timestamp: Column name containing the timestamp (optional)

        Returns:
            self: For method chaining

        Raises:
            ValueError: If a source has already been set
        """
        if self.source is not None:
            raise ValueError(
                "Source already set. from_dataset() or from_stream() can only be called once."
            )
        self.source = source
        self.source_type = FeatureSourceType.STREAM
        self.select_config = SelectConfig(
            entity=entity,
            values=values,
            timestamp=timestamp,
        )
        return self

    def aggregate(
        self,
        function: AggregateFunction,
        windows: List[timedelta],
    ) -> "FeatureBuilder":
        """
        Configure time-windowed aggregation for this feature.

        This method must be called after from_dataset() or from_stream() to specify
        aggregation over the selected columns.

        Args:
            function: The aggregation function (COUNT, MEAN, SUM, MIN, MAX, etc.)
            windows: List of time windows to aggregate over

        Returns:
            self: For method chaining

        Raises:
            ValueError: If from_dataset() or from_stream() has not been called first,
                       if aggregate() has already been called, or if no windows specified
        """
        if self.select_config is None:
            raise ValueError(
                "from_dataset() or from_stream() must be called before aggregate(). "
                "Use .from_dataset(source, entity=..., values=..., timestamp=...).aggregate(...)"
            )
        if self.aggregate_config is not None:
            raise ValueError("aggregate() can only be called once per FeatureBuilder.")
        if not windows:
            raise ValueError("At least one window must be specified")
        self.aggregate_config = AggregateConfig(
            function=function,
            windows=windows,
        )
        return self

    def backfill_from(
        self,
        source: Any,
        entity: str,
        values: str,
        timestamp: str,
    ) -> "FeatureBuilder":
        """
        Configure backfill from a batch source (for stream features).

        This is used to populate historical data for stream features
        from an existing batch dataset with potentially different column names.

        Args:
            source: The batch source to backfill from
            entity: Entity column name in the backfill source
            values: Value column name in the backfill source
            timestamp: Timestamp column name in the backfill source

        Returns:
            self: For method chaining

        Raises:
            ValueError: If not a stream feature, or if backfill already configured
        """
        if self.source_type != FeatureSourceType.STREAM:
            raise ValueError("backfill_from() can only be used with stream features")
        if self.backfill_config is not None:
            raise ValueError(
                "backfill_from() can only be called once per FeatureBuilder."
            )
        self.backfill_config = BackfillConfig(
            source=source,
            entity=entity,
            values=values,
            timestamp=timestamp,
        )
        return self

    def name(self, name: str) -> "FeatureBuilder":
        """
        Set the base name for this feature.

        This method allows manually specifying the feature name instead of
        deriving it from the entity attribute name.

        Args:
            name: The base name for the feature

        Returns:
            self: For method chaining

        Example:
            ```python
            feature = (
                ff.Feature()
                .name("custom_feature_name")
                .variant("v1")
                .from_dataset(source, entity="user_id", values="amount", timestamp="ts")
            )
            ```
        """
        self.base_name = name
        return self

    def variant(self, variant: str) -> "FeatureBuilder":
        """
        Set the initial/proposed variant for this feature.

        This method allows manually specifying the feature variant instead of
        using an auto-generated or default variant. The value provided here is
        considered "initial" or "proposed" because it may be changed during the
        registration process.

        **Equivalence Resolution Behavior:**

        During `client.apply()` (which calls `create_all()`), the server performs
        equivalence resolution to detect if an existing feature with matching
        semantics already exists. This affects the final variant used:

        - **If an equivalent feature is found:** The `FeatureVariant.variant` will
          be updated to match the existing equivalent variant, overriding the value
          specified here. This ensures semantic deduplication across the system.

        - **If no equivalent is found:** The variant specified here will be used
          as-is when creating the new feature.

        This means that calling `.variant("my_v1")` does not guarantee the final
        feature will have variant `"my_v1"` - it sets the proposed variant that
        will be used if no equivalent exists.

        Args:
            variant: The proposed variant string for the feature

        Returns:
            self: For method chaining

        Example:
            ```python
            feature = (
                ff.Feature()
                .name("transaction_count")
                .variant("production_v2")  # Proposed variant; may change if equivalent exists
                .from_dataset(source, entity="user_id", values="amount", timestamp="ts")
            )
            ```
        """
        self.initial_variant = variant
        return self

    # =========================================================================
    # Resource Interface Methods
    # =========================================================================

    def name_variant(self) -> Tuple[str, str]:
        """Return the (name, variant) tuple for this feature."""
        if self.base_name is None:
            raise ValueError("Feature base_name not set")
        if self.initial_variant is None:
            raise ValueError("Feature initial_variant not set")
        return (self.base_name, self.initial_variant)

    def get_resource_type(self) -> ResourceType:
        """Return the resource type for this feature."""
        return ResourceType.FEATURE_VARIANT

    def to_key(self) -> Tuple[ResourceType, str, str]:
        """Return the unique key for this resource."""
        if self.base_name is None:
            raise ValueError("Feature base_name not set")
        if self.initial_variant is None:
            raise ValueError("Feature initial_variant not set")
        return self.get_resource_type(), self.base_name, self.initial_variant

    def _has_aggregation(self) -> bool:
        """Check if this builder has aggregation configured."""
        return self.aggregate_config is not None

    def _is_stream(self) -> bool:
        """Check if this is a stream feature."""
        return self.source_type == FeatureSourceType.STREAM

    def get_windows(self) -> List[timedelta]:
        """Get the list of configured windows, or an empty list if not aggregated."""
        if self.aggregate_config is None:
            return []
        return self.aggregate_config.windows

    # =========================================================================
    # Build Method
    # =========================================================================

    def build(self) -> BuiltFeatures:
        """
        Build and return BuiltFeatures containing FeatureVariant(s).

        This method validates the configuration and creates the appropriate
        FeatureVariant resources. For aggregated features with multiple windows,
        one FeatureVariant is created per window.

        Each FeatureVariant has its own variant field that can be updated
        during equivalence resolution.

        Note: Before calling build(), the following fields should be normalized:
        - entity: Should be a string (entity name), not an EntityRegistrar
        - owner: Should have a default value if not set
        - value_type: Should be a ScalarType, not None or string

        Returns:
            BuiltFeatures: Wrapper containing FeatureVariant(s)

        Raises:
            ValueError: If required configuration is missing
        """
        # Validate required fields
        if self.source is None:
            raise ValueError(
                "Feature source not set. Use from_dataset() or from_stream()"
            )
        if self.select_config is None:
            raise ValueError(
                "Feature columns not set. "
                "Provide entity, values, and timestamp parameters to from_dataset() or from_stream()"
            )
        if self.base_name is None:
            raise ValueError("Feature base_name not set")

        # Use pre-normalized values from builder
        source_name_variant = self._get_source_name_variant()

        # Get column configuration
        entity_column = self.select_config.entity
        value_column = self.select_config.values
        timestamp_column = self.select_config.timestamp

        has_aggregation = self._has_aggregation()

        if has_aggregation:
            # Create one FeatureVariant per window
            features: Dict[timedelta, FeatureVariant] = {}
            for window in self.aggregate_config.windows:
                window_suffix = format_window_suffix(window)
                feature_name = f"{self.base_name}_{window_suffix}"

                aggregate_def = AggregateFeature(
                    name=feature_name,
                    input_column=value_column,
                    input_type=self.value_type,
                    function=self.aggregate_config.function,
                    time_window=window,
                )

                feature_variant = FeatureVariant(
                    created=None,
                    name=feature_name,
                    variant=self.initial_variant,
                    source=source_name_variant,
                    value_type=self.value_type,
                    entity=self.entity,
                    owner=self.owner,
                    provider="",
                    description=self.description,
                    schedule="",
                    location=ResourceColumnMapping(
                        entity=entity_column,
                        value=value_column,
                        timestamp=timestamp_column,
                    ),
                    tags=self.tags,
                    properties=self.properties,
                    feature_definition=aggregate_def,
                )
                features[window] = feature_variant

            built_features: BuiltFeatures = AggregateBuiltFeatures(
                features=features,
                base_name=self.base_name,
                base_variant=self.initial_variant,
            )
        else:
            # Create a single attribute feature
            attribute_def = AttributeFeature(
                name=self.base_name,
                input_column=value_column,
                input_type=self.value_type,
            )

            feature_variant = FeatureVariant(
                created=None,
                name=self.base_name,
                variant=self.initial_variant,
                source=source_name_variant,
                value_type=self.value_type,
                entity=self.entity,
                owner=self.owner,
                provider="",
                description=self.description,
                schedule="",
                location=ResourceColumnMapping(
                    entity=entity_column,
                    value=value_column,
                    timestamp=timestamp_column,
                ),
                tags=self.tags,
                properties=self.properties,
                feature_definition=attribute_def,
            )

            built_features = AttributeBuiltFeature(
                feature=feature_variant,
                base_name=self.base_name,
                base_variant=self.initial_variant,
            )

        # Store reference for bracket indexing access
        self._built_features = built_features

        return built_features

    # =========================================================================
    # Registration Methods
    # =========================================================================

    def register(self) -> None:
        """
        Register this feature with the global registrar.

        This method is called by the @ff.entity decorator to register
        the feature with the Featureform backend.
        """
        from . import global_registrar

        global_registrar.register_feature_builder(self)

    # =========================================================================
    # Helper Methods
    # =========================================================================

    def _get_source_name_variant(self) -> Tuple[str, str]:
        """Get the (name, variant) tuple from the source."""
        if hasattr(self.source, "name_variant"):
            return self.source.name_variant()
        elif isinstance(self.source, tuple) and len(self.source) == 2:
            return self.source
        else:
            raise ValueError(
                f"Source must have a 'name_variant()' method or be a (name, variant) tuple. "
                f"Got: {type(self.source).__name__}"
            )


# =============================================================================
# Feature Factory Function
# =============================================================================


def Feature(*args, **kwargs):
    """
    Create a Feature definition.

    This is the main entry point for defining features. It supports both
    the legacy API and the new builder pattern.

    **Legacy API** (pass transformation_args as first positional argument):
        ```python
        @ff.entity
        class User:
            avg_transactions = ff.Feature(
                transformation[["CustomerID", "Amount", "Timestamp"]],
                type=ff.Float32,
                variant="quickstart",
            )
        ```

    **Builder Pattern** (no arguments, use method chaining):
        ```python
        @ff.entity
        class User:
            account_balance = (
                ff.Feature()
                .from_dataset(
                    batch_features,
                    entity="CustomerID",
                    values="Balance",
                    timestamp="Timestamp",
                )
            )

            transaction_count = (
                ff.Feature()
                .from_dataset(
                    batch_features,
                    entity="CustomerID",
                    values="Amount",
                    timestamp="Timestamp",
                )
                .aggregate(
                    function=ff.AggregateFunction.COUNT,
                    windows=[timedelta(days=1), timedelta(days=7)]
                )
            )
        ```

    Both patterns can be used together in the same entity class, allowing
    for seamless migration from the legacy API to the builder pattern.

    Returns:
        FeatureBuilder (no args) or FeatureColumnResource (with args)
    """
    if args or kwargs:
        # Legacy API - delegate to FeatureColumnResource
        from .column_resources import FeatureColumnResource

        return FeatureColumnResource(*args, **kwargs)
    else:
        # Builder Pattern - return new FeatureBuilder
        return FeatureBuilder()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Streaming provider wrapper classes for Featureform.
"""

from typing import TYPE_CHECKING, List, Optional, Union

from ..resources import KafkaTopic

if TYPE_CHECKING:
    from ..registrar import UserRegistrar

from .offline import OfflineProvider

__all__ = ["KafkaProvider"]


class KafkaProvider(OfflineProvider):
    def __init__(self, registrar, provider):
        super().__init__(registrar, provider)
        self.__registrar = registrar
        self.__provider = provider

    def register_kafka_topic(
        self,
        name: str,
        topic: str,
        description: str = "",
        tags: Optional[List[str]] = None,
        properties: Optional[dict] = None,
        owner: Union[str, "UserRegistrar"] = "",
    ):
        """Register a Kafka topic as a primary data source.

        **Examples**

        ```
        spark = client.get_provider("my_spark")
        transactions_stream = spark.register_kafka_topic(
            name="transactions_stream",
            topic="transactions",
            description="A stream of transaction data from Kafka"
        )
        ```

        Args:
            name (str): Name to register the Kafka topic under.
            topic (str): The name of the Kafka topic.
            variant (str, optional): Variant name, if applicable.
            owner (Union[str, "UserRegistrar"], optional): Owner of the data source.
            description (str, optional): Description of the Kafka topic.
            tags (List[str], optional): Tags associated with the data source.
            properties (dict, optional): Additional properties.

        Returns:
            source (ColumnSourceRegistrar): The registered data source.
        """

        tags, properties = set_tags_properties(tags, properties)

        kafka_topic = KafkaTopic(
            name=name,
            topic=topic,
            provider=self.name(),
            owner=owner,
            description=description,
        )
        self.__registrar.add_resource(kafka_topic)
        return kafka_topic
